#ifndef DEBUG_H
#define DEBUG_H

#include <stdio.h>

/* NOTE:  In order to use these macros the int variables debugLevel and
 * debugMask must be defined (usually in main() ).
 */

/*
 *                      Debug Support
 *                       DVM 6/22/92
 *
 * This include file provides several debugging capabilities.
 *
 * o Print statements and/or other debugging actions may be
 *   easily enabled or disabled by the DEBUG flag at compile
 *   time.
 *
 * o Debug actions may optionally be controlled at run time using
 *   two global variables:
 *
 *   debugLevel - Each piece of debug code is assigned a level at
 *                compile time. At run time, its value is
 *                compared to the value of debugLevel. The code
 *                is executed only if debugLevel is greater than
 *                or equal to its value. A negative value of
 *                debugLevel, e.g., -N, enable ONLY debug code at
 *                level N.
 *
 *                Debug levels must be positive. They are
 *                provided as parameters to the debugging macro.
 *
 *   debugMask  - Each piece of debug code is also associated with
 *                a 32-bit value which has one or more bits set
 *                in it. At run time, this value is ANDed
 *                (bitwise) with the value of debugMask. If the
 *                result is non-zero, the code is executed. This
 *                allows each task, for instance to be assigned
 *                its own bit. At run time, each task's debug
 *                output may be enabled or disabled.
 *
 *                Debug bits are provided through the following
 *                macros:
 *
 *                DEBUG_DIR - This gets defined in sw/target.mk. It
 *                  is set to a different value for each directory
 *                  (at least it will be after all directories are
 *                  implemented.). It should be a 32-bit constant
 *                  with only one bit set in it - see the list
 *                  below in this file. If it is not defined,
 *                  either the DEBUG_FILE definition is used, or
 *                  DEBUG_MISC is used.
 *
 *                DEBUG_FILE - If a file is to have its own value
 *                  for this, it should be defined in the .mk file
 *                  in that file's directory. See sw/target.mk for
 *                  an example of how to do this. This may be
 *                  applied to multiple files if desired. NOTE that
 *                  if DEBUG_DIR is also defined, these two bits
 *                  are OR-ed together to form the debug bits. This
 *                  means that if either of the two bits are set in
 *                  debugMask at run time, the debug statement(s)
 *                  will be executed. 
 *
 *                Note that this file serves as a repository of
 *                the different values for these debug constants
 *                (see the bottom of this file).
 *
 * Including this file makes available all of the following:
 *
 * Macro            Action/Definition
 * =========================================================
 * ifmdebug         This acts just like an if statement which
 *                  does not require any arguments in
 *                  parentheses. It means that the statement(s)
 *                  which follow will be executed only if the
 *                  debugMask bits match the value of DEBUGBITS
 *                  set at compile time.  It means,
 *                  'if-masked-debug'.  Note that the actions are
 *                  not associated with a debugLevel, and will be
 *                  taken no matter what the value of debugLevel.
 *
 *         example: ifmdebug {
 *                      filesystems_mounted++;
 *                      strcpy(last_fs_mounted,"This file system");
 *                  }
 *
 * ifmldebug(dlev)    This acts just like an if statement which is
 *                  just as above, except it also associates a
 *                  level with the actions. That is, the actions
 *                  will be taken if the DEBUGBITS/debugMask test
 *                  allows it AND if the level is allowed by the
 *                  debugLevel value.
 *
 *         example: ifmldebug(ThisLevel) {
 *                      last_fs_mount_time=gettime();
 *                  }
 *
 * DPRINTF          This acts just like printf(), except that if
 *                  DEBUG is not defined, it will do nothing.
 *                  Note that this prints independently from the
 *                  value of debugMask or debugLevel.
 *
 *         example: DPRINTF("Activating IM\n");
 *
 * DMPRINTF         This acts just like printf(), but it only
 *                  occurs if DEBUGBITS/debugMask allow it. Note
 *                  that it is independent of debugLevel.
 *
 *         example: DMPRINTF("IM statistics:\n");
 *                  DMPRINTF("  Command count = %d.\n",im_cmd_count);
 *
 * DMLPRINTF(dlev)  This acts just like printf(), except that it
 *                  only occurs if the DEBUGBITS/debugMask allow
 *                  it AND if the debugLevel allows it. The dlev
 *                  argument sets the level.
 *
 *                  DMLPRINTF(oPrmLev)("ReadCap returned : 0x%x.\n",err); 
 *
 */



static int noprintf(char * s1,...)
{
    return 0;
}

#define DPRINTF noprintf
#define ifmdebug    if(0)
#define ifmldebug(x)  if(0)


#define DMPRINTF ifmdebug DPRINTF
#define DMLPRINTF(dlev) ifmldebug(dlev) DPRINTF

/* Define debug bit masks */

#define DEBUG_MISC       0x00000001
#define DEBUG_INITR      0x00000002
#define DEBUG_IM         0x00000004
#define DEBUG_QMGR       0x00000008
#define DEBUG_DCP        0x00000010
#define DEBUG_OSA        0x00000020
#define DEBUG_OSAUTIL    0x00000040
#define DEBUG_VENEER     0x00000080
#define DEBUG_OSYS       0x00000100
#define DEBUG_GENHOST    0x00000200
#define DEBUG_MYSTIC     0x00000400
#define DEBUG_DMA        0x00000800
#define DEBUG_DIAG       0x00001000
#define DEBUG_HW         0x00002000
#define DEBUG_XFR        0x00004000

#define trace  DPRINTF
#define trace1 DPRINTF
#define trace2 DPRINTF

#endif /* DEBUG_H */

