/********************************************************************/
/*  o_ad.c - Functions to manipulate Allocation Descriptors         */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrCore/o_ad.c,v $
 $Revision: 1.23 $      $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:52:50 $

 ********************************************************************/

#include "../nsrHdrs/def.h"
#include "Debug.h"
#include "../nsrHdrs/o_vndops.h"
#include "../nsrHdrs/nsr.h"
#include "../nsrHdrs/o_node.h"
#include "../nsrHdrs/o_proto.h"
#include "../nsrHdrs/vfs.h"
#include "../nsrHdrs/nsrerrno.h"

/* utility routines */

/*
 *  ot_numADs(otp)
 *      returns the number of allocation descriptors used to
 *      represent this file
 *
 *  input
 *      The otnode for the file
 *
 *  returns
 *      The integral number of allocation descriptors in the ADL
 *      associated with the otnode.
 *
 */
UINT32 ot_numADs(struct otnode *otp)
{
    return(otp->ADL.endADs - otp->ADL.beginADs);
}

/*
 * ot_ICBHasADs(otp)
 *
 * This returns TRUE if and only if the ICB associated with the supplied
 * otnode should contain ADs. (NOTE that only FEs and USEs return TRUE.)
 */
UINT32 ot_ICBHasADs(struct otnode *otp) {
    return(otp->icbe.ge.sTag.uTagID == TAGID_FILE_ENTRY ||
           otp->icbe.ge.sTag.uTagID == TAGID_UNALLOC_SP_ENTRY);
}

/*
 * ot_copyADs()
 *
 * This copies Allocation Descriptors, between an array of Generic ADs
 * and an area of memory starting at Dest L_AD long containing an array
 * of Short, Long, or Generic ADs. It copies in either direction, and
 * performs all the conversions required. Its operation is limited by
 * L_AD as well as endADp.
 *
 * The "Destination" contains the Short, Long, or Extended ADs to be
 * converted FROM or TO, while the AD array always contains Generic ADs
 * only. The Generic AD array is to be used by the code, while the Dest
 * area is to be used on the disk. This is possible because Generic ADs
 * hold a superset of the information in the other two AD types.
 *
 * Control parameters:
 *
 *   dir     - FROM_OP copies from the destination to the array
 *             TO_OP   copies to the destination from the array
 *   ADType  - This sets the type of Allocation Descriptors at the
 *             Destination.
 *             ADSHORT, ADLONG, or ADEXTENDED
 *   L_AD    - This is the total number of bytes at the Destination which
 *             are available for Allocation Descriptors.
 *
 * Destination parameters
 *
 *   Destpp  - address of pointer to Destination. At invocation, this
 *             supplies the starting location. When it returns, it
 *             contains the next address to be written to.
 *
 *             TO   - This returns with a pointer to the location into
 *                    which the Allocation Extent AD must go.
 *
 *             FROM - This returns with a pointer to the location in
 *                    which the terminating AD was found (see notes on
 *                    ADpp below), or the end of supplied space, L_AD
 *                    bytes after the supplied Destpp location.
 *
 * Allocation Descriptor array parameters:
 *
 *   ADpp    - pointer to pointer to next AD. When called, this supplies
 *             starting point. When returns, supplies next point at which
 *             to continue.
 *
 *             NOTE that when copying FROM the Dest, it will point to the
 *             terminating entry in the AD. This can be:
 *             - The AD for the Allocation Extent
 *             - The AD with the ExtentLength == 0
 *             - The same address as endADp
 *             (see ECMA 167 P4-12)
 *
 *   endADp  - pointer to one beyond last gen_AD entry
 *
 * This returns an error code of zero if successful. Other possible
 * errors are:
 *
 *   XXXXXX  - Xxxxx xxxx x xxxxx x xxxxxxxxx.
 */
int ot_copyADs(UINT16 dir,
               UINT16 ADType,
               UINT32 L_AD,
               void **Destpp,
               struct gen_AD **ADpp,
               struct gen_AD *endADp)
{
    struct short_AD *sADp      = *Destpp;
    struct long_AD  *lADp      = *Destpp;
    struct ext_AD   *xADp      = *Destpp;
    UINT16          SizeOfAD   = ot_sizeOfAD(ADType);
    void            *endDestp  = (char *)*Destpp + L_AD;
    void            *lastDestp = (char *)endDestp - SizeOfAD;

    /* First, we hold the size of the AD and change from the the L_AD */
    /* domain over to the domain of the last location to which an */
    /* entire AD may be copied (lastDest). This allows the test for */
    /* whether we can copy to be at the beginning of the loop, without */
    /* risking exceeding the supplied limit.  */

    if (dir == TO_OP) {
        while (*Destpp <= lastDestp && *ADpp + 1 <= endADp) {
            switch (ADType) {
            case ADSHORT:
                sADp->ExtentLength = (*ADpp)->ExtentLength;
                sADp->Location     = (*ADpp)->Location_LBN;
                *Destpp = ++sADp;
                break;
            case ADLONG:
                lADp->ExtentLength    = (*ADpp)->ExtentLength;
                lADp->Location_LBN    = (*ADpp)->Location_LBN;
                lADp->Location_PartNo = (*ADpp)->Location_PartNo;
                lADp->aImpUse[0]      = (*ADpp)->aImpUse[0];
                lADp->aImpUse[1]      = (*ADpp)->aImpUse[1];
                memset(&lADp->aImpUse[2],0,4); /* fill others with zeroes */
                *Destpp = ++lADp;
                break;
            case ADEXTENDED:
            default:
                memmove(xADp,*ADpp,SizeOfAD);
                *Destpp = ++xADp;
            }
            (*ADpp)++;
        }
        if (*ADpp != endADp) {
            /* Back up to point to last entry */
            *Destpp = lastDestp;
            (*ADpp)--;
        }
        /* Fill all remaining space with nulls */
        if (*Destpp < endDestp) {
            memset(*Destpp,0,(int)((char *)endDestp-(char *)*Destpp));
        }
    } else {
        while (*Destpp <= lastDestp &&
               *ADpp + 1 <= endADp) {
            memset((*ADpp)->aImpUse,0,6);
            switch (ADType) {
            case ADSHORT:
                (*ADpp)->RecordedLength =
                    (*ADpp)->InfoLength = sADp->ExtentLength.bf.Length;
                (*ADpp)->ExtentLength   = sADp->ExtentLength;
                (*ADpp)->Location_LBN   = sADp->Location;
                *Destpp = ++sADp;
                break;
            case ADLONG:
                (*ADpp)->RecordedLength  =
                    (*ADpp)->InfoLength  = lADp->ExtentLength.bf.Length;
                (*ADpp)->ExtentLength    = lADp->ExtentLength;
                (*ADpp)->Location_LBN    = lADp->Location_LBN;
                (*ADpp)->Location_PartNo = lADp->Location_PartNo;
                (*ADpp)->aImpUse[0]      = lADp->aImpUse[0];
                (*ADpp)->aImpUse[1]      = lADp->aImpUse[1];
                *Destpp = ++lADp;
                break;
            case ADEXTENDED:
            default:
                memmove(*ADpp,xADp,SizeOfAD);
                *Destpp = ++xADp;
            }
            if ((*ADpp)->ExtentLength.Length32 == 0 ||
                (*ADpp)->ExtentLength.bf.Type == E_ALLOCEXTENT) {
                /* Back up to point to last entry */
                /* NOTE that ADpp is not yet incremented */
                *Destpp = (char *)*Destpp - SizeOfAD;
                break;
            }
            (*ADpp)++;
        }
    }
    return(0);
}

/*
 * ot_setADType()
 *
 * This function walks through the AD array associated with the supplied
 * otnode, and sets the appropriate AD Type flags in the file entry.
 *
 * ISSUE: It currently only supports Long and Extended ADs.
 *
 * It returns the ADType it sets in the file entry.
 */
UINT16 ot_setADType(struct otnode *otp)
{
    UINT16         ADT;
    struct gen_AD *ADp;
    
    for (ADp = otp->ADL.beginADs,ADT = ADLONG;
         ADp < otp->ADL.endADs && ADT == ADLONG;
         ADp++) {
        if (ADp->ExtentLength.bf.Length != ADp->RecordedLength
            || ADp->RecordedLength != ADp->InfoLength) {
            ADT=ADEXTENDED;
        }
    }
    otp->icbe.ge.sICBTag.Flags &= ~ADTYPEMASK;  /* clear ADTYPE field */
    otp->icbe.ge.sICBTag.Flags |= ADT;
    return(ADT);
}

/*
 * ot_ensureAESpace(otp,ADType,srcADp)
 *
 * This function makes sure that enough space exists in the supplied
 * otnode's ADL diskSpace linked list to complete the writing of all the
 * ADs which remain in the ADL, starting with the one at the supplied
 * srcADp.
 *
 * Unneeded space is freed.
 *
 * Note that it may return with ENOSPC, which means that there is not
 * enough room on the disk for the required allocation extents.
 *
 */
int ot_ensureAESpace(struct otnode *otp,
                     UINT16 ADType,
                     struct gen_AD *srcADp) {

    struct space *dsp;
    struct space *nextdsp;
    struct otfs  *fsp    = (struct otfs *) otp->vnode.v_vfsp->vfs_data;
    int          error   = 0;
    INT32        numADs  = otp->ADL.endADs - srcADp;
    UINT32       used;
    UINT32       AEBlocks;

    /* If WORM, need to kill all the space and the space list. */
    if (fsp->MediaType <= (int) fsWORM) { /* WORM */
        if (otp->ADL.diskSpace) {
            /* WORM: Kill all space and spaceList */
            killSpaceList(&otp->ADL.diskSpace);
        }
    }
    /* First, walk through all of the space elements in the list, */
    /* keeping track of how many ADs remain to be written. */
    dsp = nextdsp = otp->ADL.diskSpace;
    while (nextdsp) {
        dsp = nextdsp;
        numADs -= ((fsp->SectorSize*dsp->length - AE_HdrLen)
                   / ot_sizeOfAD(ADType));
        if (numADs > 0) {
            numADs ++;  /* used for Allocation Extent AD */
        } else {
            if (numADs < 0) {
                /* NOTE that numADs is negative, so this will take */
                /* away unused sectors... */
                used = dsp->length + (numADs*ot_sizeOfAD(ADType)/fsp->SectorSize);
                numADs = 0;
                if ((error = returnSectors(fsp, (struct space *)0,
                                           dsp->start + used,
                                           dsp->length - used,
                                           bmapErased)) != 0) {
                    return(error);
                }
                dsp->length = used;
            }
            /* Then, return the rest of the list. */
            if (dsp->next) {
                if((error = returnSectors(fsp, dsp->next, 0, 0,
                                          bmapErased)) != 0) {
                    return(error);
                }
                killSpaceList(&dsp->next);
            }
        }
        nextdsp = dsp->next;
    }
    /* If there is no spaceList, dsp is NULL. Otherwise, it is */
    /* pointing to the last valid space structure in the list. */
    if (numADs > 0) {
        /* Still have some ADs to make room for. */
        AEBlocks = ((numADs*ot_sizeOfAD(ADType)
                     + AE_HdrLen) / fsp->SectorSize) + 1;
        nextdsp = ot_getSpace(fsp,AEBlocks,contiguous);
        if (!nextdsp) {
            error = ENOSPC;
        } else {
            if (!dsp) {         /* Replaces empty list */
                otp->ADL.diskSpace = nextdsp;
            } else {            /* Appends to list */
                dsp->next = nextdsp;
                nextdsp->previous = dsp;
            }
        }
    }
    return(error);
}

/*
 * ot_putAEAD()
 *
 * This forms an AD corresponding to the Allocation Extent described by
 * the space structure pointed to by dsp, and copies it to the address
 * defined by destADp.
 */
int ot_putAEAD(struct otfs *fsp,
               void *destADp,
               UINT16 ADType,
               struct space *dsp)
{
    struct gen_AD AEAD;
    struct gen_AD *AEADp = &AEAD;
    int error;
    
    AEAD.ExtentLength.bf.Type = E_ALLOCEXTENT;
    AEAD.InfoLength =
        AEAD.RecordedLength =
            AEAD.ExtentLength.bf.Length = dsp->length * fsp->SectorSize;
    AEAD.Location_LBN    = dsp->start;
    AEAD.Location_PartNo = 0;
    memset(AEAD.aImpUse,0,6);
    error = ot_copyADs(TO_OP,ADType,
                       ot_sizeOfAD(ADType),
                       &destADp,&AEADp,AEADp+1);
    return(error);
}

/*
 * ot_initAE()
 *
 * This initializes an Allocation Extent which starts at the address
 * supplied in the destpp. The AE itself is defined by the AEsize and
 * AELoc parameters. The L_AD field is calculated, filled in, and
 * returned. Note that the prevAELoc parameter provides the backwards
 * reference for the AE.
 *
 * When this returns, destpp points to a pointer to the location of the
 * first AD to fit into the AE.
 */
UINT32 ot_initAE(UINT32 AEsize,
                 UINT32 AELoc,
                 UINT32 prevAELoc,
                 UINT16 ADType,
                 void **destpp)
{
    struct AllocationExtentDesc *AEp = *destpp;
    void                        *unusedLoc;
    int                          unusedLen;

    FillTag(TAGID_ALLOC_EXTENT,1,1,NOCRC,0,AELoc,&AEp->sTag);
    *destpp = AE_ADStart(AEp);
    AEp->prevAllocExtLoc = prevAELoc;
    AEp->L_AD  = AEsize - AE_HdrLen;
    unusedLen  = AEp->L_AD % ot_sizeOfAD(ADType);
    AEp->L_AD -= unusedLen;
    unusedLoc = (char *)*destpp + AEp->L_AD;
    memset(unusedLoc,0,unusedLen); /* zero unused bytes */
    return(AEp->L_AD);
}

/*
 * ot_recordedBytes()
 *
 * This takes the supplied ADL and returns the number of bytes which have
 * been recorded inside it.
 */
UINT32 ot_recordedBytes(struct ADList *ADLp)
{
    struct gen_AD *ADp;
    UINT32 total;
    
    for (ADp = ADLp->beginADs,total=0;
         ADp < ADLp->endADs;
         ADp++) {
        total += ADp->RecordedLength;
    }
    return(total);
}

/*
 * ot_enforceInfoLength()
 *
 * This takes the supplied otnode and enforces the rule that the
 * sum of the InfoLengths of the extents must equal the InfoLength of the
 * file. This is done by trusting that the InfoLength for the file is
 * correct, and thus simply reducing the last extent's InfoLength if
 * required.
 */
void ot_enforceInfoLength(struct otnode *otp)
{
    struct otfs   *fsp   = (struct otfs *) otp->vnode.v_vfsp->vfs_data;
    struct ADList *ADLp  = &otp->ADL;
    UINT32         total = otp->icbe.fe.InfoLengthL;
    UINT32         unused;
    struct gen_AD *ADp;
    
    for (ADp = ADLp->beginADs;
         ADp < ADLp->endADs;
         ADp++) {
        if (total < ADp->InfoLength) {
            ADp->InfoLength = total;
        }
        total -= ADp->InfoLength;
    }
}

/*
 * ot_writeADL()
 *
 * This function performs any writing of the allocation descriptor lists
 * associated with the supplied otnode as required. It should be called
 * just before writing the icbe, as the first portion of the Allocation
 * Descriptor List resides in the icbe.
 *
 * Overview of algorithm:
 *
 * if <ADL has been modified> {
 *     if <FE> {
 *         <Set LogBlocks field in FE>;
 *         <enforce InfoLength sum == total>;
 *     }
 *     <set type of ADs to be used>;
 *     <prepare FE for ADs>;
 *     <copy ADs into FE/USE>
 *     <ensure space for remaining ADs>
 *     if <more ADs to be written> {
 *         <put AD for next AE>
 *         <find largest AE>
 *         <allocate buffer to hold largest AE>
 *         while <more ADs to be written> {
 *             <initialize AE>
 *             <copy ADs into AE>
 *             if <more ADs to be written> {
 *                 <put AD for next AE>
 *             }
 *             <translate AE to disk format>
 *             <write AE to disk>
 *         }
 *         <free AE buf>
 *         <deallocate list of space used for AEs>
 *     }
 * }
 */
int ot_writeADL(struct otnode *otp)
{
    UINT32         L_AD;
    UINT16         ADType;
    int            error = 0;
    void           *destp;
    UINT32         AEBufLen;
    void           *AEBuf;
    struct gen_AD  *srcADp = otp->ADL.beginADs;
    struct otfs    *fsp    = (struct otfs *) otp->vnode.v_vfsp->vfs_data;
    struct space   *dsp;
    struct space   *tdsp;
    UINT32         prevLoc;
    int            unusedLen;
    void           *unusedLoc;

    struct ADList *ADLp  = &otp->ADL;
    
    if (ADLp->modified) {
        if (otp->icbe.ge.sTag.uTagID == TAGID_FILE_ENTRY) {
            otp->icbe.fe.LogBlocksH = 0;
            otp->icbe.fe.LogBlocksL = ot_recordedBytes(ADLp)/fsp->SectorSize;
            ot_enforceInfoLength(otp);
        }
        
        ADType = ot_setADType(otp);

        /* First, fill in all primary ADs in icbe. */

        switch (otp->icbe.ge.sTag.uTagID) {
        case TAGID_FILE_ENTRY:
            destp     = FE_ADStart(&otp->icbe.fe);
            L_AD      = fsp->SectorSize - FE_HdrLen(&otp->icbe.fe);
            unusedLen = L_AD % ot_sizeOfAD(ADType);
            L_AD     -= unusedLen;
            unusedLoc = (char *)destp + L_AD;
            otp->icbe.fe.L_AD = L_AD;
            break;
        case TAGID_UNALLOC_SP_ENTRY:
            destp     = USE_ADStart(&otp->icbe.fe);
            L_AD      = fsp->SectorSize - USE_HdrLen;
            unusedLen = L_AD % ot_sizeOfAD(ADType);
            L_AD     -= unusedLen;
            unusedLoc = (char *)destp + L_AD;
            otp->icbe.uspe.L_AD = L_AD;
        }
        memset(unusedLoc,0,unusedLen); /* zero unused bytes */

        otd_printADs("writeADL(ADL)",ADGENERIC,
                     (char *)otp->ADL.endADs - (char *)otp->ADL.beginADs,
                     otp->ADL.beginADs);
        
        if (error = ot_copyADs(TO_OP,ADType,L_AD,
                               &destp,&srcADp,
                               otp->ADL.endADs))
            return (error);

        if (!error) {
            switch (otp->icbe.ge.sTag.uTagID) {
            case TAGID_FILE_ENTRY:
                otd_printADs("writeADL(FE)",ADType,L_AD,FE_ADStart(&otp->icbe.fe));
                break;
            case TAGID_UNALLOC_SP_ENTRY:
                otd_printADs("writeADL(USE)",ADType,L_AD,USE_ADStart(&otp->icbe.uspe));
            }
        }
        
        /* Make sure we have diskSpace avail for all AEs */
        if (error = ot_ensureAESpace(otp,ADType,srcADp))
            return(error);

        if (srcADp < otp->ADL.endADs) {

            /* We DO have some AEs to write... */
            dsp   = otp->ADL.diskSpace;
            if (error = ot_putAEAD(fsp,destp,ADType,dsp))
                return(error);
            /* Allocate a buffer to hold the largest AE. */
            AEBufLen=0;
            for (tdsp=dsp;tdsp;tdsp=tdsp->next) {
                if (AEBufLen < tdsp->length)
                    AEBufLen = tdsp->length;
            }
            AEBufLen *= fsp->SectorSize;
            AEBuf = kmem_alloc(AEBufLen,BUFRAM);
            if (AEBuf == 0)
                return(ENOMEM);
            
            prevLoc = otp->icb_addr_LBN;

            while (srcADp < otp->ADL.endADs) {

                destp = AEBuf;
                L_AD = ot_initAE(dsp->length*fsp->SectorSize,
                                 dsp->start,prevLoc,ADType,
                                 &destp);
                if (error = ot_copyADs(TO_OP,ADType,L_AD,
                                       &destp,&srcADp,
                                       otp->ADL.endADs))
                    break;

                otd_printADs("writeADL(AE)",ADType,L_AD,AE_ADStart(AEBuf));

                if (srcADp < otp->ADL.endADs) {
                    if (error = ot_putAEAD(fsp,destp,ADType,dsp->next))
                        break;
                }
#if !IBM
                ot_xlate(TAGID_ALLOC_EXTENT,TO_OP,
                         (char *)AEBuf,
                         dsp->length*fsp->SectorSize,
                         ADType);
#endif          
                if (error = ot_write(fsp,dsp->start,AEBuf,
                                     dsp->length*fsp->SectorSize))
                    break;
                prevLoc = dsp->start;
                dsp     = dsp->next;
            }
            kmem_free(AEBuf,AEBufLen,BUFRAM);
            if (error == 0) {
                kmem_free(otp->ADL.beginADs,otp->ADL.bufLen,BUFRAM);
                killSpaceList(&otp->ADL.diskSpace);
                memset(&otp->ADL,0,sizeof(struct ADList));
            }
        }
    }
    return(error);
}

/*
 * ot_roundUp()
 *
 * This rounds up the supplied length to the nearest sector boundary
 * according to the supplied sector size.
 *
 * NOTE: This is used ONLY to be able to read MDI's NSR disks, where the
 *       last extent is not a multiple of a sector size.
 */
UINT32 ot_roundUp(UINT32 sectorSize,UINT32 length)
{
    UINT32 partial = length % sectorSize;
    
    if (partial != 0) {
        length += sectorSize - partial;
    }
    return(length);
}

/*
 * ot_readADL()
 *
 * This reads in the entire set of ADs for a given otnode. These are
 * placed in the ADL structure in the otnode.
 *
 * It first copies the ADs from of the FileEntry, and then reads in all
 * the AEs in order, and copies all of their entries into the ADL.
 *
 * if <valid ADType> {
 *     <malloc ADL buffer>
 *     <copy ADs from FE/USE>
 *     while <next AD is to AE> {
 *         <make sure AE buffer is big enough (and even exists...)>
 *         <read AE into AE buffer>
 *         <add AE space to ADL spaceList> ((WORM??))
 *         <translate structure from disk>
 *         <make sure ADL buffer is big enough (use realloc())>
 *         <copy ADs from AE>
 *     }
 *     <free AE buffer>
 * }
 */
int ot_readADL(struct otnode *otp)
{
    int          error = 0;
    struct otfs  *fsp = (struct otfs *)otp->vnode.v_vfsp->vfs_data;

    void *AEBuf   = NULL;
    int  AEBufLen = 0; /* allocated size */
    int  AELen    = 0; /* valid bytes in AEBuf */
    
    /* Temporaries for reallocating space */
    int           ADLMaxEnts;
    int           ADLEnts;
    int           AEMaxEnts;
    int           newADLBufLen;
    UINT32        newBufLen;

    struct ADList *ADLp  = &otp->ADL;
    const int     MinADBufLen = (((fsp->SectorSize/sizeof(struct short_AD))+1)
                                 * sizeof(struct gen_AD));
    const UINT16  ADType = otp->icbe.ge.sICBTag.Flags & ADTYPEMASK;
    struct space  *dsp;

    void          *destp;       /* temp */
    UINT32        length;       /* temp */

    if (!ValidADType(ADType)) {
        error = EINVAL;
    } else {
        /* First, initialize the ADL structure and allocate the ADL buffer */
        ADLp->beginADs = ADLp->endADs =
            (struct gen_AD *)kmem_alloc(MinADBufLen,BUFRAM);
        ADLp->diskSpace = NULL;
        ADLp->bufLen = MinADBufLen;
        ADLp->modified = 0;

        /* Next, copy all ADs from FE into ADL */
        switch (otp->icbe.ge.sTag.uTagID) {
        case TAGID_FILE_ENTRY:
            destp  = FE_ADStart(&otp->icbe.fe);
            length = MIN(otp->icbe.fe.L_AD,fsp->SectorSize);
            otd_printADs("readADL(FE)",ADType,length,destp);
            break;
        case TAGID_UNALLOC_SP_ENTRY:
            destp  = USE_ADStart(&otp->icbe.fe);
            length = MIN(otp->icbe.uspe.L_AD,fsp->SectorSize);
            otd_printADs("readADL(USE)",ADType,length,destp);
        }
        if (error = ot_copyADs(FROM_OP,ADType,
                               length,&destp,
                               &ADLp->endADs,
                               ADBufLim(ADLp))) {
            /* ISSUE - what about deallocating the ADL buffer? */
            /*       - Should it not be done here? */
            return(error);
        }

        /*
         * NOTE the following possible states at this point:
         *   1. *ADpp reached ADBufLim:
         *      - This should NEVER happen. We prevent this in our
         *        calculations of MinADBufLen.
         *   2. Terminating entry found at destination.
         *      - The entries preceding it are all the ADs.
         *   3. destpp reached *destpp + length:
         *      - This is a proper termination of the ADs.
         *      - EOF is interpreted as not continuing beyond the entries
         *        in the ICBE.
         *   4. AE entry found:
         *      - The ADs continue inside the AE.
         */

        while (ADLp->endADs->ExtentLength.bf.Type == E_ALLOCEXTENT) {
            AELen = ADLp->endADs->ExtentLength.bf.Length;
            if (AEBufLen < AELen) {
                /* Make sure AE buffer is big enough */
                AEBuf = kmem_realloc(AEBuf,AELen,AEBufLen,BUFRAM);
                if (!AEBuf) {
                    error    = ENOMEM;
                    AEBufLen = 0;
                    AELen    = 0;
                    break;
                } else {
                    AEBufLen = AELen;
                }
            }
            
            /* Read AE into AEBuf */
            length = AELen;       /* MAY change readLen */
            if (error=ot_read(fsp,ADLp->endADs->Location_LBN,AEBuf,&length))
                break;

            /* Add AE space into ADLp->diskSpace */
            dsp = ADLp->diskSpace; /* MAY change dsp... */
            if (error = ot_addToList(ADLp->endADs->Location_LBN,
                                     AELen/fsp->SectorSize,
                                     &dsp,fsp))
                break;
#if !IBM
            /* Translate structure from disk. */
            ot_xlate(TAGID_ALLOC_EXTENT,
                     FROM_OP,(char *)AEBuf,
                     AELen,ADType);
#endif
            if (error = ot_goodTag(TAGID_ALLOC_EXTENT,AEBuf,
                                   ADLp->endADs->Location_LBN))
                break;

            /* Make sure ADL buffer is big enough */
            ADLMaxEnts = ADLp->bufLen/sizeof(struct gen_AD);
            ADLEnts    = ADLp->endADs - ADLp->beginADs;
            AEMaxEnts  = AELen/ot_sizeOfAD(ADType);
            
            if (ADLEnts + AEMaxEnts > ADLMaxEnts) {
                newBufLen = ((ADLEnts + AEMaxEnts)
                             * sizeof(struct gen_AD));
                ADLp->beginADs = kmem_realloc(ADLp->beginADs,
                                              newBufLen, ADLp->bufLen,
                                              CPURAM);
                if (!ADLp->beginADs) {
                    error = ENOMEM;
                    ADLp->endADs = ADLp->beginADs;
                    ADLp->bufLen = 0;
                    break;
                } else {
                    ADLp->endADs = ADLp->beginADs + ADLEnts;
                    ADLp->bufLen = newBufLen;
                }
            }
            /* copy ADs from AE */
            length = MIN(AELen - AE_HdrLen,
                         ((struct AllocationExtentDesc *)AEBuf)->L_AD);
            destp = AE_ADStart(AEBuf);

            otd_printADs("readADL(AE)",ADType,length,destp);

            if (error = ot_copyADs(FROM_OP,ADType,
                                   length,&destp,
                                   &ADLp->endADs,
                                   ADBufLim(ADLp)))
                break;
        }
        kmem_free(AEBuf,AEBufLen,BUFRAM);
    }

    /* Round final extent to multiple of sector size */
    if (!error) {
        if (ADLp->endADs-1 >= ADLp->beginADs) {
            (ADLp->endADs-1)->ExtentLength.bf.Length =
                ot_roundUp(fsp->SectorSize,(ADLp->endADs-1)->ExtentLength.bf.Length);
            (ADLp->endADs-1)->RecordedLength =
                ot_roundUp(fsp->SectorSize,(ADLp->endADs-1)->RecordedLength);
            (ADLp->endADs-1)->InfoLength =
                ot_roundUp(fsp->SectorSize,(ADLp->endADs-1)->InfoLength);
        }
    }
    if (!error) {
        otd_printADs("readADL(ADL)",ADGENERIC,
                     (char *)ADLp->endADs - (char *)ADLp->beginADs,
                     ADLp->beginADs);
    }
    return(error);
}

/*
 *  ot_addToADL
 *      updates or adds to a file's ADL
 *
 *  input parameters:
 *      otp:            the otnode for the file
 *      startblk:       logical sector offset into file at which
 *                      this change begins
 *      extType:        indicates whether or not the space
 *                      described by addrList is E_RECORDED or
 *                      E_ALLOCATED
 *      addrList:       a list of space entries to be used for this
 *                      modification.
 *  output parameters:
 *      otp:            its ADL has been modified.
 *  returns:
 *      0               success
 *      ENOMEM:         a memory allocation failed
 *
 * BASIC OUTLINE
 * 1. Search through existing ADL to find the ADs holding the beginning
 *    and the end of the changes.
 * 2. Prepend allocated space onto addrList if appending beyond the end
 *    of the file's existing ADL space.
 * 3. Determine changes to the beginning and ending ADs if required. The
 *    first part of the beginning AD and/or the last part of the ending
 *    AD  may need to be kept as is, requiring changes to their locations
 *    and lengths.
 * 4. Insert or delete the appropriate number of ADs in the ADL.
 * 5. Make change to the beginning AD.
 * 6. Copy addrList into ADL.
 * 7. Make change to ending AD.
 * 8. Kill prepended portion of addrList.
 * 9. Mark otnode as modified.
 */
INT32 ot_addToADL(struct otnode *otp,
                  UINT32 startBlk,
                  UINT32 extType,
                  struct space *addrList)
{
    int            append;      /* temporary boolean */

    UINT32         newExtType;  /* used to set extType in new ADL ents */
    UINT16         numToWrite;  /* accumulates number of ADs to write */
    UINT16         numToDelete; /* Holds how many to be overwritten or deleted */
    UINT32         endBlk;      /* sector offset into file */
                                /* at which this change ends. */
    UINT32         iIndex;      /* See below in code. */
    UINT32         jIndex;      /* See below in code. */
    UINT32         iBegExt;     /* See below in code. */
    UINT32         jBegExt;     /* See below in code. */
    int            iFound;      /* See below in code. */
    int            jFound;      /* See below in code. */

    struct gen_AD  borderADs[2];/* Holding place for */
    struct space   *orgAddrs;   /* gets new spacelist & original */
                                /* spacelist loc.*/
    struct otfs    *fsp = (struct otfs *) otp->vnode.v_vfsp->vfs_data;
    struct space   *addrPtr;    /* temp spacelist pointer */
    struct gen_AD  *ADp;        /* temp AD ptr */
    UINT32         ADIndex;     /* temp AD Index */

    INT32          error;

    /* First, we derive endBlk and numToWrite from addrList and */
    /* startBlk by walking through addrList */

    for (addrPtr = addrList, endBlk = startBlk, numToWrite = 0;
         addrPtr;
         endBlk += addrPtr->length, addrPtr = addrPtr->next, numToWrite++);

    /* Zero out some local vars. */
    error = 0;
    memset(borderADs,0,(2*sizeof(struct gen_AD)));

    /* find ADs corresponding to begin/end-Extent: */

    iFound  = ot_searchADL(otp,startBlk,&iBegExt,&iIndex);
    jFound  = ot_searchADL(otp,endBlk-1,&jBegExt,&jIndex);

    DMLPRINTF(oADLev)("ot_addToADL: End ADL scans: iBegExt=0x%x, jBegExt=0x%x\n",
                      iBegExt,jBegExt);
    
    /* Now, the following variables hold this information:
     *   iBegExt - Offset into file to beginning of extent
     *             containing beginning of modifications
     *   jBegExt - Offset into file to beginning of extent
     *             containing last block to be modified.
     *   iIndex  - Index into ADL of extent containing beginning of
     *             modifications.
     *   jIndex  - Index into ADL of extent containing last block to
     *             be modified.
     *   iFound  - iIndex is valid.
     *   jFound  - jIndex is valid.
     */

    /* add to addrList if necessary and calculate borderADs */
    append = (!iFound && !jFound); /* ISSUE: Do we need 'append'? */
    orgAddrs = 0;
    if (append) {               /* appending! */
        /* Insert UNRECORDED ADs before new ones if there's a gap */
        /* between our start and the old end. */
        /* (jBegExt == First block after end of ADs when !jFound.) */
        if (jBegExt < startBlk) {
            /* prepend new spacelist to addrList */         
            orgAddrs = ot_getSpace(fsp,startBlk-jBegExt,nonContiguous);
            if (orgAddrs) {
                numToWrite++;
                for (addrPtr = orgAddrs;
                     addrPtr->next;
                     addrPtr = addrPtr->next, numToWrite++);
                addrPtr->next = addrList;
                addrList->previous = addrPtr;
                addrList = orgAddrs;
                orgAddrs = addrPtr->next;   /* (old addrList) */
            } else {
                error = ENOMEM;
            }
        }
        /* and leave borderADs zeroed */
    } else {                        /* NOT appending */
        /* NOTE that these calculations may allow the length fields to */
        /* be set to zeroes, in which case, the borderAD is not */
        /* needed. If not zero, borderAD[0] should overwrite extent at */
        /* iIndex, and borderAD[1] should overwrite extent at jIndex. */
        if (iFound) {
            /* borderADs[0] : old data (first part of old I extent) */
            borderADs[0] = *(otp->ADL.beginADs+iIndex);
            borderADs[0].InfoLength =
                borderADs[0].RecordedLength =
                    borderADs[0].ExtentLength.bf.Length =
                        (startBlk-iBegExt) * fsp->SectorSize;
        }
        if (jFound) {
            /* borderADs[1] : old data (last part of old J extent) */
            borderADs[1] = *(otp->ADL.beginADs+jIndex);
            borderADs[1].InfoLength =
                borderADs[1].RecordedLength =
                    borderADs[1].ExtentLength.bf.Length -=
                        (endBlk-jBegExt) * fsp->SectorSize;
            borderADs[1].Location_LBN += endBlk-jBegExt;
        }
    }

    /* delete ADs or insert space for new ones as necessary */
    if (!error) {
        numToDelete = append ? 0 : jIndex - iIndex + 1;
        if (borderADs[0].ExtentLength.bf.Length > 0)
            numToWrite++;
        if (borderADs[1].ExtentLength.bf.Length > 0)
            numToWrite++;
        /*
         * numToWrite is TOTAL number of extents to write, including:
         *   number of extents passed in
         *   number to fill gap when appending beyond EOF
         *   number of borderADs
         * numToDelete is number of old entries no longer needed:
         *   All extents between I and J inclusive.
         */
        ADIndex = iIndex;
        if (!append && borderADs[0].ExtentLength.bf.Length > 0) {
            ADIndex++;
        }
        /* ADIndex is the insertion point: beyond borderADs[0] */
        if (numToWrite > numToDelete) {
            error = ot_insertADs(otp,ADIndex,numToWrite-numToDelete);
        } else if (numToWrite < numToDelete) {
            error = ot_deleteADs(otp,ADIndex,numToDelete-numToWrite);
        }
    }
    /* actually write any new entries: */

    if (!error) {
        ADp = otp->ADL.beginADs+iIndex;
        /* Write in borderADs[0] */
        if (borderADs[0].ExtentLength.bf.Length > 0) {
            *ADp++ = borderADs[0];
        }
        /* Write in addrList */
        /* Set newExtType to unrecorded if filling gap. */
        newExtType = (orgAddrs == 0 ? extType : E_ALLOCATED);
        for (addrPtr = addrList; addrPtr; addrPtr = addrPtr->next) {
            if (addrPtr == orgAddrs) /* Done with filling gap. */
                newExtType = extType;
            DMLPRINTF(oADLev)("Adding: st=%x, len=%x.\n",
                              addrPtr->start,  addrPtr->length);
            ADp->Location_LBN = addrPtr->start;
            ADp->Location_PartNo = 0;
            memset(ADp->aImpUse,0,6);
            ADp->ExtentLength.bf.Length =
                ADp->RecordedLength =
                    ADp->InfoLength = addrPtr->length * fsp->SectorSize;
            ADp->ExtentLength.bf.Type = newExtType;
            otd_printADs("Added:",ADGENERIC,sizeof(struct gen_AD),ADp);
            ADp++;
        }
        /* Write in borderADs[1] */
        if (borderADs[1].ExtentLength.bf.Length > 0) {
            *ADp++ = borderADs[1];
        }
    }
    
    /* clean up: */

    if (orgAddrs) { /* we prepended some stuff to addrList sent in; free it */
        orgAddrs->previous->next = 0;       /* terminate addrList */
        orgAddrs->previous = 0;             /* axe prev in first entry */
        killSpaceList(&addrList);           /* NOW kill prepended */
                                            /* entries (does NOT */
                                            /* return to freelist). */
    }
    if (!error) {
        otp->ADL.modified = TRUE;
        otp->flags |= OT_MOD;
#ifdef USECONSOLIDATE
        error = ot_consolidateADL(otp);
#endif
    }
    return error;
} /* end ot_addToADL */

/*
 *  ot_searchADL()
 *
 *      function to scan the ADL for an entry containing a given logical
 *      block for a file. It starts searching at the beginning of the ADL
 *      and walks through it, keeping track of the offset into the file
 *      corresponding to each extent as it goes. It stops when it finds
 *      the desired location, or when it reaches the end of the ADL.
 *
 *  input parameters:
 *
 *      otp:        The otnode for the file.
 *      startblk:   The logical block in the file to search for
 *
 *  output parameters:
 *
 *      begExtentp: The logical sector number for the entry
 *      ADIndexp:   The entry offset into the ADL for the entry
 *                  which contains startblk.
 *
 *      NOTE:       To get a pointer to the AD which terminated the
 *                  search, simply add the ADIndex to the beginning of
 *                  the ADs, as shown below:
 *
 *                  ADp = (otp->ADL.beginADp + *ADIndexp).
 *  returns:                                                        
 *      1           found
 *      0           At end of list without finding it.
 */
INT32 ot_searchADL(struct otnode *otp,
                   UINT32 startBlk,
                   UINT32 *begExtp,
                   UINT32 *ADIndexp)
{
    struct otfs   *fsp = (struct otfs *) otp->vnode.v_vfsp->vfs_data;
    int           found = 0;
    struct gen_AD *ADp = otp->ADL.beginADs;

    *ADIndexp = 0;
    *begExtp  = 0;
    while (ADp < otp->ADL.endADs) {
        if (startBlk >= *begExtp &&
            startBlk <= *begExtp+ADp->InfoLength/fsp->SectorSize-1) {
            found = 1;
            break;
        }
        *begExtp += ADp->InfoLength/fsp->SectorSize;
        ADp++;
        (*ADIndexp)++;
    }
    return(found);
} /* end ot_searchADL */

#ifdef USECONSOLIDATE
/*
 *  ot_consolidateADL
 *      routine to reduce unneccessary ADs
 *
 *  input parameters:
 *      otp:    the otnode for the file whose ADL is to be
 *              consolidated
 *  returns:
 *      0       success
 */
INT32 ot_consolidateADL(struct otnode *otp)
{
    struct gen_AD  *iADp;
    struct gen_AD  *jADp;
    struct otfs    *fsp = (struct otfs *) otp->vnode.v_vfsp->vfs_data;
    INT32          error;
    UINT32         removedCount;   /* number of entries consolidated */
    
    error = 0;
    iADp = otp->ADL.beginADs;
    while (!error && iADp + 1 < otp->ADL.endADs) {
        jADp = iADp + 1;
        while ((iADp->Location_LBN
                + iADp->ExtentLength.bf.Length/fsp->SectorSize == jADp->Location_LBN) &&
               (iADp->Location_PartNo == jADp->Location_PartNo) &&
               iADp->ExtentLength.bf.Type == jADp->ExtentLength.bf.Type) {
            iADp->InfoLength =
                iADp->RecordedLength =
                    iADp->ExtentLength.bf.Length +=
                        jADp->ExtentLength.bf.Length;
            jADp++;
        }
        removedCount = jADp - iADp - 1;
        if (removedCount) {
            error = ot_deleteADs(otp,iADp-otp->ADL.beginADs + 1,removedCount);
        }
        iADp++;
    }
    return error;
} /* end ot_consolidateADL */
#endif

/*
 *  ot_deleteADs
 *      routine to delete one or more entries from an ADL
 *
 *  input parameters:
 *      otp:    the otnode for the file whose ADL is to be modified
 *      offset: offset into ADL, in terms of entries, where the
 *              deletion should begin
 *      count:  the number of entries to delete
 *
 *  returns:
 *      0       success
 *
 *  notes:
 *      offset is 0-based.
 */
INT32 ot_deleteADs(struct otnode *otp,
                   UINT32 offset,
                   UINT32 count)
    
{
    struct gen_AD  *deleteADp;  /* point to begin deletion */

    deleteADp = otp->ADL.beginADs + offset;
    memmove(deleteADp,
            deleteADp+count,
            (otp->ADL.endADs-deleteADp) * sizeof(struct gen_AD));
    memset(deleteADp+count,0,count*sizeof(struct gen_AD));
    otp->ADL.endADs -= count;
    otp->ADL.modified = TRUE;
    return(0);
} /* end ot_deleteADs */

/*
 *  ot_insertADs
 *      insert space for one or more ADs into ADL
 *
 *  input parameters:
 *      otp:        The otnode for the data file whose ADL is to
 *                  be modified
 *      offset:     The 0-based entry offset into the ADL where the
 *                  insertion should begin. This should be within the
 *                  bounds of the file.
 *      count:      The number of entries to insert. This must be <= half
 *                  the number of entries which will fit into a sector.
 *
 *  returns:
 *      0           success
 *      ENOMEM:     memory allocation failure
 */
INT32 ot_insertADs(struct otnode *otp,
                   UINT32 offset,
                   UINT32 count)
{
    struct gen_AD  *newBeginADs;
    struct gen_AD  *insertADp;  /* point to begin insertion */
    const UINT32   ExtraBufGrowth = 4 * sizeof(struct gen_AD);
    
    UINT32 error  = 0;
    UINT32 currentNumADs = ot_numADs(otp);
    UINT32 bufReq = (currentNumADs + count) * sizeof(struct gen_AD);

    if (bufReq > otp->ADL.bufLen) {
        /* Grow the ADL buffer. */
        bufReq += ExtraBufGrowth;
        newBeginADs = kmem_realloc(otp->ADL.beginADs,
                                   bufReq,
                                   otp->ADL.bufLen,CPURAM);
        if (!newBeginADs) {
            error = ENOMEM;
        } else {
            otp->ADL.endADs   = newBeginADs + currentNumADs;
            otp->ADL.beginADs = newBeginADs;
            memset(otp->ADL.endADs,0,bufReq-otp->ADL.bufLen);
            otp->ADL.bufLen = bufReq;
        }
    }
    if (!error) {
        insertADp = otp->ADL.beginADs + offset;
        memmove(insertADp+count,
                insertADp,
                (otp->ADL.endADs-insertADp) * sizeof(struct gen_AD));
        memset(insertADp,0,count*sizeof(struct gen_AD));
        otp->ADL.endADs += count;
        otp->ADL.modified = TRUE;
    }
    return(error);
} /* end ot_insertADs */

void otd_printADs(char *msg,
                  UINT16 ADType,
                  UINT32 length,
                  void *ptr)
{
    int numADs = length / ot_sizeOfAD(ADType);
    int i;
    char ETypeStr[4] = {'R','A','U','X'};
    
    struct short_AD *sadp = ptr;
    struct long_AD  *ladp = ptr;
    struct ext_AD   *xadp = ptr;
    struct gen_AD   *gadp = ptr;
    
    char *iu2fmt = "IU= 0x%2.2x 0x%2.2x\n";
    char *iu6fmt = "IU= 0x%2.2x 0x%2.2x 0x%2.2x 0x%2.2x 0x%2.2x 0x%2.2x\n";
    char *sfmt   = "[%c] EL=0x%x, Loc=0x%x\n";
    char *lfmt   = "[%c] EL=0x%x, BN=0x%x, PN=0x%x, ";
    char *xfmt   = "[%c] EL=0x%x, RL=0x%x, IL=0x%x, BN=0x%x, PN=0x%x, ";
    
    if (numADs > 0) {
        if (numADs == 1) {
            DMLPRINTF(oADLev)("<%s> ",msg);
        } else {
            DMLPRINTF(oADLev)("-----<%s>-----\n",msg);
        }
        
        switch (ADType)
        {
        case ADSHORT:
            for( i=0; i< numADs; i++) 
            {
                DMLPRINTF(oADLev)(sfmt,ETypeStr[sadp->ExtentLength.bf.Type],
                                  sadp->ExtentLength.bf.Length,sadp->Location);
                sadp++;
            }
            break;
        case ADLONG:
            for( i=0; i< numADs; i++) 
            {
                DMLPRINTF(oADLev)(lfmt,ETypeStr[ladp->ExtentLength.bf.Type],
                                  ladp->ExtentLength.bf.Length,
                                  ladp->Location_LBN,ladp->Location_PartNo);
                DMLPRINTF(oADLev)(iu6fmt,
                                  ladp->aImpUse[0],ladp->aImpUse[1],
                                  ladp->aImpUse[2],ladp->aImpUse[3],
                                  ladp->aImpUse[4],ladp->aImpUse[5]);
                ladp++;
            }
            break;
        case ADEXTENDED:
            for( i=0; i< numADs; i++) 
            {
                DMLPRINTF(oADLev)(xfmt,ETypeStr[xadp->ExtentLength.bf.Type],
                                  xadp->ExtentLength.bf.Length,
                                  xadp->RecordedLength,xadp->InfoLength,
                                  xadp->Location_LBN,xadp->Location_PartNo);
                DMLPRINTF(oADLev)(iu2fmt,
                                  xadp->aImpUse[0],xadp->aImpUse[1]);
                xadp++;
            }
            break;
        caseADGENERIC:
            for( i=0; i< numADs; i++) 
            {
                DMLPRINTF(oADLev)(xfmt,msg,ETypeStr[gadp->ExtentLength.bf.Type],
                                  gadp->ExtentLength.bf.Length,
                                  gadp->RecordedLength,gadp->InfoLength,
                                  gadp->Location_LBN,gadp->Location_PartNo);
                DMLPRINTF(oADLev)(iu6fmt,
                                  gadp->aImpUse[0],gadp->aImpUse[1],
                                  gadp->aImpUse[2],gadp->aImpUse[3],
                                  gadp->aImpUse[4],gadp->aImpUse[5]);
                gadp++;
            }
            break;
        }
    }
}

/*
 * ot_truncateADL()
 *
 * This truncates the supplied otnode's ADL at the point corresponding to
 * the supplied ADIndex. The file's size, etc. are not affected, but all
 * ADs starting with the one at ADIndex are removed from the file, and
 * the disk space they take up is returned to the appropriate space
 * bitmaps.
 *
 * The otnode's ADL is marked as modified, and any errors (from attempts
 * to return the space) are returned.
 *
 */
INT32 ot_truncateADL(struct otnode *otp,UINT32 ADIndex)
{
    struct otfs   *fsp = (struct otfs *) otp->vnode.v_vfsp->vfs_data;
    struct gen_AD *ADp;
    INT32         error = 0;
    
    /* WORM extents can't be recovered */
    if (fsp->MediaType > (int) fsWORM) {
        /* delete data described by ADs */
        for (ADp = otp->ADL.beginADs + ADIndex;
             ADp < otp->ADL.endADs && !error;
             ADp++) {
            if (ADp->ExtentLength.bf.Type == E_RECORDED) {
                error = returnSectors(fsp,(struct space *) 0,
                                      ADp->Location_LBN,
                                      ADp->ExtentLength.bf.Length/fsp->SectorSize,
                                      bmapUnerased);
            } else if (ADp->ExtentLength.bf.Type == E_ALLOCATED) {
                error = returnSectors(fsp,(struct space *) 0,
                                      ADp->Location_LBN,
                                      ADp->ExtentLength.bf.Length/fsp->SectorSize,
                                      bmapErased);
            }
        }
    }
    otp->ADL.endADs = otp->ADL.beginADs + ADIndex;
    otp->ADL.modified = TRUE;
    otp->flags |= OT_MOD;
    return error;
}


