/********************************************************************/
/*  o_struct.c - NSR structure related routines                     */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/*  Module name:    o_struct.c                                      */
/*                                                                  */
/*  Description:                                                    */
/*      Routines to translate disk and in-memory structures         */
/*      Routines to get/put 16-bit and 32-bit values                */
/*      Routines to swap bytes of 16-bit and 32-bit values          */
/*      Routines to copy strings                                    */
/*      Routines to swap basic structures (tag, extent_ad)          */
/*                                                                  */
/*  Functions in this module:                                       */
/*      ot_xlate        getint      putint      ot_copy             */
/*      ot_xlateBlock   getuint     putlong     xlate_tag           */
/*      xlate_extent    getlong     swapUINT16  xlate_ICBTag        */
/*      FillTag         getulong    swapUINT32  xlate_timestamp     */
/*      isTag                                                       */
/*                                                                  */
/*  Modification history:                                           */
/*      16.Mar.92   ECB     Changed to ANSI function declarations   */
/*      01.Jul.92   ECB     Added support for ECMA 167 structs      */
/*      01.Dec.92   ECB     FillTag added                           */
/*                                                                  */
/*                                                                  */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrCore/o_struct.c,v $
 $Revision: 1.46 $      $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:52:59 $
 ********************************************************************/

#include "def.h"
#include <sys/types.h>
#include <string.h>
#include "o_vndops.h"
#include "Debug.h"
#include "nsr.h"
#include "o_proto.h"

#if SUN             /* to enable debugging */
#include "o_fs.h"
#endif

/* internal use functions forward declarations */

void xlate_tag( struct Tag *x );
void xlate_ICBTag( struct ICBTag *x, UINT32 y );
void xlate_domainID( struct domainEntityId *deip );
void xlate_udfID( struct udfEntityId *ueip );
void xlate_implID( struct implEntityId *ieip );
void xlate_extent( struct extent_ad *x );
void xlate_short_AD( struct short_AD *lap );
void xlate_long_AD( struct long_AD *lap );
void xlate_ext_AD( struct ext_AD *eap );
void xlate_ADs( void *ADp, UINT32 ADType, UINT32 L_AD );
void xlate_timestamp( struct timestamp *tsp );
void xlate_fid(struct FileIDDesc *fidp,INT32 nbytes,UINT32 direction );

#ifdef DO_CRC
static unsigned short cksum(unsigned char *s,register int n);
#endif

extern INT32 otfs_debug;

/*
 * ot_sizeOfAD()
 *
 * This function returns the number of bytes required by the supplied
 * ADType, which can be ADSHORT, ADLONG, or ADEXTENDED.
 */
UINT32 ot_sizeOfAD(UINT16 ADType) {
    if        (ADType == ADSHORT) {
        return (sizeof(struct short_AD));
    } else if (ADType == ADLONG) {
        return (sizeof(struct long_AD));
    } else if (ADType == ADEXTENDED) {
        return (sizeof(struct ext_AD));
    } else if (ADType == ADGENERIC) {
        return (sizeof(struct gen_AD));
    } else { /* ISSUE: What to do with ADNONE? */
        return (1);
    }
}

/********************************************************************/
/*  ot_xlate                                                        */
/*      translate optical disk structures from/to system            */
/*      memory structures                                           */
/*                                                                  */
/*  input parameters:                                               */
/*      type:       structure type:                                 */
/*                      TAGID_XXX (see nsr.h)                       */
/*      direction:  which way to translate: FROM_OP or TO_OP        */
/*      dskstruct:  the structure to translate                      */
/*      fidBytesToXlate: number of bytes in the fid to xlate        */
/*                       (Can't xlate past end of buffer)           */
/*                                                                  */
/*  output parameters:                                              */
/*      dskstruct:  contains the translated structure               */
/*                                                                  */
/*  returns:                                                        */
/*      0           success                                         */
/*      EINVAL:     invalid structure                               */
/*                                                                  */
/*  notes:                                                          */
/*      the value of "fidBytesToXlate" is only important when       */
/*          translating file identifiers which are variable length. */
/*      return value of EINVAL probably means one of two things:    */
/*          the structure type does not match the type requested;   */
/*          or, the structure is corrupt                            */
/*      the structure is translated in situ                         */
/*                                                                  */
/*  Modification history:                                           */
/*                                                                  */
/********************************************************************/
#if !IBM
void ot_xlate(int type,
              INT32 direction,
              register caddr_t dskstruct,
              UINT32 maxBytes,
              UINT32 ADType)
{
    /* maxBytes is the max bytes to translate. This is only used when
       translating FIDs, LVDs, LVIDs, AEs, FEs, and USEs. The latter
       three require it to prevent translating ADs till the end of
       memory should the L_AD field be corrupted. Note that FEs and
       USEs will ALWAYS be equal to the sector size. */
    /* ADType is used when translating AEs ONLY. */

    register UINT32             i;
    struct Tag                  xlated_tag;
    struct VolDescPtr           *vdp;
    struct UnallocSpDesHead     *usdp;
    struct extent_ad            *exp;
    struct Terminator           *tp;
    struct AnchorVolDesPtr      *ap;
    struct PrimaryVolDes        *pvdp;
    struct PartDesc             *pdp;
    struct LogVolDesc           *lvdp;
    struct ImpUseDesc           *iudp;
    struct LVInformation        *lvip;
    struct FileIDDesc           *fidp;
    UCHAR8                      *maxfidp;
    union  icbe                 *fep;
    struct IndirectEntry        *idep;
    struct FileSetDesc          *fsdp;
    struct PartIntegrityEntry   *pip;
    struct UnallocSpEntry       *sptbl;
    struct SpaceBitmapEntry     *bmap;
    struct TerminalEntry        *tep;
    UINT32                      imax;
    struct PartHeaderDesc       *PHDp;
    struct AllocationExtentDesc *AEDp;
    struct LogicalVolumeIntegrityDesc *LVIDp;
    struct LVIDImplUse          *LVIDIUp;

    /*
    * Clause to handle a structure with an unknown tag.
    * The code that reads a volume descriptor sequence
    * needs to be able to do this.
    */
    if ( type == TAGID_NONE ) {
        if ( direction == FROM_OP ) {
            xlated_tag = *(struct Tag *)dskstruct;
            xlate_tag( &(xlated_tag));
            type = xlated_tag.uTagID;
        } else {
            type = ((struct Tag *)dskstruct)->uTagID;
        }
    }

    switch(type) {
        /* begin ECMA 167 structures */
        
        
        /* volume descriptor pointer */
    case TAGID_POINTER:
        vdp = (struct VolDescPtr *) dskstruct;
#ifdef DO_CRC
        /*
         * This is only shown to be an example of where and how one might 
         * do crc's.
         */
        if ( direction == TO_OP ) {
            vdp->sTag.uCRCLen = sizeof(struct VolDescPtr)-sizeof(struct Tag);
            vdp->sTag.uDescriptorCRC = cksum((unsigned char *)dskstruct+sizeof(struct Tag),
                                             vdp->sTag.uCRCLen);
        }
#endif
        xlate_tag( &(vdp->sTag));
        swapUINT32( &(vdp->uVolDescSeqNum));
        xlate_extent( &(vdp->sNextVDS));
        break;
        
        /* partition integrity entry */
    case TAGID_PART_INTEGRITY:
        pip = (struct PartIntegrityEntry *) dskstruct;
        xlate_tag( &(pip->sTag));
        xlate_timestamp(&pip->sRecordingTime);
        break;
        
        /* space bitmap entry */
    case TAGID_SPACE_BMAP:
        bmap = (struct SpaceBitmapEntry *)dskstruct;
        xlate_tag(&bmap->hdr.sTag);
        swapUINT32(&bmap->hdr.N_Bits);
        swapUINT32(&bmap->hdr.N_Bytes);
        break;
        
        /* space table entry */
    case TAGID_UNALLOC_SP_ENTRY:
        sptbl = (struct UnallocSpEntry *)dskstruct;
        xlate_tag(&sptbl->sTag);
        if( direction == FROM_OP)  {
            swapUINT32(&sptbl->L_AD);
            xlate_ICBTag(&sptbl->sICBTag, direction);
        }
        maxBytes = MIN(maxBytes-USE_HdrLen,sptbl->L_AD);
        xlate_ADs(USE_ADStart(sptbl),
                  sptbl->sICBTag.Flags & ADTYPEMASK,
                  maxBytes);
        if ( direction == TO_OP ) {
            swapUINT32(&sptbl->L_AD);
            xlate_ICBTag(&sptbl->sICBTag, direction);
        }
        break;
        
        /* unallocated volume space descriptor */
    case TAGID_USD:
        usdp = (struct UnallocSpDesHead *) dskstruct;
        xlate_tag( &(usdp->sTag));
        swapUINT32( &(usdp->uVolDescSeqNum));
        
        /* now point to the beginning of the Space records, which are extent_ad's */
        exp = (struct extent_ad *) (((UCHAR8 *) usdp) + sizeof( struct UnallocSpDesHead));
        if( direction == FROM_OP) {
            swapUINT32( &(usdp->uNumAllocationDes));
        }
        for(i=0; i<usdp->uNumAllocationDes; i++) {
            xlate_extent( exp);
            exp += sizeof( struct extent_ad);
        }
        if( direction == TO_OP) {
            swapUINT32( &(usdp->uNumAllocationDes));
        }
        break;
        
        /* terminating descriptor */
    case TAGID_TERM_DESC:
        tp = (struct Terminator *) dskstruct;
        xlate_tag( &(tp->sTag));
        break;
        
    case TAGID_LVID:
        LVIDp = (struct LogicalVolumeIntegrityDesc *) dskstruct;
        xlate_tag( &(LVIDp->sTag));
        xlate_timestamp(&LVIDp->sRecordingTime);
        swapUINT32( &(LVIDp->integrityType));
        xlate_extent( &(LVIDp->nextIntegrityExtent));
        swapUINT64(&(LVIDp->UniqueIdH),&(LVIDp->UniqueIdL));
        /* Free Space and Size Tables */
        if (direction == FROM_OP) {
            swapUINT32( &(LVIDp->N_P));
            swapUINT32( &(LVIDp->L_IU));
        }
        /* recalculate maxBytes to mean max bytes left to xlate */
        maxBytes -= LVID_HdrLen;
        {
            UINT32 maxN_U32 = maxBytes / sizeof(UINT32);

            UINT32 maxFst = MIN(maxN_U32,(LVIDp->N_P));
            UINT32 maxSzt;      /* easier to calculate later */

            /* Translate all free space table entries */
            for ( i=0; i<maxFst; i++ ) {
                swapUINT32( LVID_FSTStart(LVIDp)+i );
            }
            
            /* Calculate how many U32s are left, and the max number of
               size table entries */
            maxN_U32 = (maxN_U32 > maxFst)? maxN_U32 - maxFst : 0; 
            maxSzt = MIN(maxN_U32,LVIDp->N_P);

            /* Translate all size table entries */ 
            for ( i=0; i<maxSzt; i++ ) {
                swapUINT32( LVID_SZTStart(LVIDp)+i );
            }
        }

        /* Map the LVIDIUp onto the implementationUse area */
        LVIDIUp = (struct LVIDImplUse *)(LVID_IUStart(LVIDp));
        
        xlate_implID( &(LVIDIUp->implementationID));
        /* xlate numFiles, if enough bytes */
        if ((char *)&(LVIDIUp->numFiles)-dskstruct > sizeof(UINT32)) {
            swapUINT32( &(LVIDIUp->numFiles));
        }
        /* xlate numDirectories, if enough bytes */
        if ((char *)&(LVIDIUp->numDirectories)-dskstruct > sizeof(UINT32)) {
            swapUINT32( &(LVIDIUp->numDirectories));
        }
        /* No xlate required for implementationUse */

        if ( direction == TO_OP ) {
            swapUINT32( &(LVIDp->N_P ));
            swapUINT32( &(LVIDp->L_IU ));
        }
        break;

        /* term entry */
    case TAGID_TERM_ENTRY:
        tep = (struct TerminalEntry *) dskstruct;
        xlate_tag( &(tep->sTag));
        xlate_ICBTag(&tep->sICBTag, direction);
        break;
        
        /* anchor pointer volume descriptor */
    case TAGID_ANCHOR:
        ap = (struct AnchorVolDesPtr *) dskstruct;
        xlate_tag( &(ap->sTag));
        xlate_extent( &(ap->sMainVDSAdr));
        xlate_extent( &(ap->sReserveVDSAdr));
        break;
        
        /* primary volume descriptor */ 
    case TAGID_PVD:
        pvdp = (struct PrimaryVolDes *) dskstruct;
        xlate_tag( &(pvdp->sTag));
        swapUINT32( &(pvdp->uVolDescSeqNum));
        swapUINT32( &(pvdp->uPrimVolDesNum));
        /* dstring, no translation */
        swapUINT16( &(pvdp->uVSN));
        swapUINT16( &(pvdp->uMaxVSN));
        swapUINT16( &(pvdp->uInterchangeLev));
        swapUINT16( &(pvdp->uMaxInterchangeLev));
        swapUINT32( &(pvdp->uCharSetList));
        swapUINT32( &(pvdp->uMaxCharSetList));
        /* dstring, no translation */
        /* charspec, no translation */
        /* charspec, no translation */
        xlate_extent( &(pvdp->sVolAbstract));
        xlate_extent( &(pvdp->sVolCopyrightNotice));
        /* ApplID regid, no xlate */
        xlate_timestamp( &(pvdp->sRecordingTime));
        xlate_implID( &(pvdp->sImplementationID));
        /* Implementation Use, no xlate */
        swapUINT32( &(pvdp->uPredecessorVDSLoc));
        swapUINT16( &(pvdp->uFlags));
        /* aReserved - no xlate */
        break;
        
        /* partition descriptor */
    case TAGID_PD:
        pdp = (struct PartDesc *) dskstruct;
        xlate_tag( &(pdp->sTag));
        swapUINT32( &(pdp->uVolDescSeqNum));
        swapUINT16( &(pdp->uPartFlags));   
        swapUINT16( &(pdp->uPartNumber));
        /* regid, no xlate */
        
        /* bytes for partition contents use - see 4/14.3 */
        PHDp = (struct PartHeaderDesc *) pdp->aPartContentsUse;
        xlate_short_AD(&PHDp->UST);
        xlate_short_AD(&PHDp->USB);
        xlate_short_AD(&PHDp->PIT);
        xlate_short_AD(&PHDp->FST);
        xlate_short_AD(&PHDp->FSB);
        swapUINT32(&pdp->uAccessType);
        swapUINT32( &(pdp->uPartStartingLoc));
        swapUINT32( &(pdp->uPartLength));
        xlate_implID( &(pdp->sImplementationID));
        /* implementation use */
        break;
        
        /* logical volume descriptor */
    case TAGID_LVD:
        lvdp = (struct LogVolDesc *) dskstruct;
        xlate_tag( &(lvdp->sTag));
        swapUINT32( &(lvdp->uVolDescSeqNum));
        /* charspec, no xlate */
        /* dstring, no xlate (logical volume identifier) */
        swapUINT32( &(lvdp->uLogBlkSize));
        xlate_domainID( &(lvdp->sDomainID));
        /* contents use field (long_AD pts to FSD) */
        xlate_long_AD((struct long_AD *)&lvdp->uLogVolUse[0]);
        if (direction == FROM_OP) {
            swapUINT32( &(lvdp->uMapTabLen));
            swapUINT32( &(lvdp->uNumPartMaps));
        }
        xlate_implID( &(lvdp->sImplementationID));
        /* bytes, no xlate */
        xlate_extent( &(lvdp->integritySeqExtent));

        /* Limit to bytes available OR mapTableLen from LVD, whichever
           is smaller. */
        maxBytes -= LVD_HdrLen;
        maxBytes = MIN(maxBytes,lvdp->uMapTabLen);
        {
            /* Limit number of table entries to as many as will fit in
               the bytes available, or the number indicated in the
               LVD, whichever is smaller. */
            UINT32 maxN_PM = maxBytes / sizeof(struct PartMap1);
            maxN_PM = MIN(maxN_PM,(lvdp->uNumPartMaps));

            /* Translate all partition map table entries */
            for ( i=0; i<maxN_PM; i++ ) {
                swapUINT16( &((LVD_PM1Start(lvdp))[i].uVSN));
                swapUINT16( &((LVD_PM1Start(lvdp))[i].uPartNum));
            }
        }
        if (direction == TO_OP) {
            swapUINT32( &(lvdp->uMapTabLen));
            swapUINT32( &(lvdp->uNumPartMaps));
        }
        break;
        
        /* implementation use volume descriptor */
    case TAGID_IUD:
        iudp = (struct ImpUseDesc *) dskstruct;
        xlate_tag( &(iudp->sTag));
        swapUINT32( &(iudp->uVolDescSeqNum));
        /* NOTE that this will translate even if NOT OSTA IUVD!, since
           the structure is defined with a udfEntityID, rather than a
           generic regid. */
        if (direction == FROM_OP) {
            xlate_udfID( &(iudp->sImplementationIdentifier));
        }
        if (strncmp((char *)iudp->sImplementationIdentifier.aID,
                    OSTA_LVINFO_ID, 23) == 0) {
            lvip = (struct LVInformation *) iudp->aReserved;
            /* charspec - no xlate */
            /* dstring - no xlate */
            /* dstring - no xlate */
            /* dstring - no xlate */
            /* dstring - no xlate */
            xlate_implID( &(lvip->sImplementationID));
            /* 128 bytes of implementation use, no xlate */
        } else {
            /* 460 bytes of implementation use, no xlate */
        }
        if (direction == TO_OP) {
            xlate_udfID( &(iudp->sImplementationIdentifier));
        }
        break;
        
        /* File Set Descriptor */
    case TAGID_FSD:
        fsdp = (struct FileSetDesc *) dskstruct;
        xlate_tag(&fsdp->sTag);
        xlate_timestamp(&fsdp->sRecordingTime);
        swapUINT16(&fsdp->uInterchangeLev);
        swapUINT16(&fsdp->uMaxInterchangeLev);
        swapUINT32(&fsdp->uCharSetList);
        swapUINT32(&fsdp->uMaxCharSetList);
        swapUINT32(&fsdp->uFileSetNum);
        swapUINT32(&fsdp->uFileSetDescNum);
        /* charspec, no xlate */
        /* dstring, no xlate */
        /* charspec, no xlate */
        /* dstring, no xlate */
        /* dstring, no xlate */
        /* dstring, no xlate */
        xlate_long_AD(&fsdp->sRootDirICB);
        xlate_domainID( &(fsdp->DomainID));
        xlate_long_AD(&fsdp->sNextExtent);
        /* 48 bytes of reserved use, no xlate */
        break;
        
        /* file identifier descriptor */
    case TAGID_FILE_ID:
        fidp = (struct FileIDDesc *) dskstruct;
        xlate_fid(fidp,maxBytes,direction);
        break;
        
    case TAGID_FILE_ENTRY:
        fep = (union  icbe *) dskstruct;
        xlate_tag(&(fep->fe.sTag));
        swapUINT32(&(fep->fe.UID));
        swapUINT32(&(fep->fe.GID));
        swapUINT32(&(fep->fe.Permissions));
        swapUINT16(&(fep->fe.LinkCount));
        /* RecFormat goes here - no xlate */
        /* RecDisplayAttr goes here - no xlate */
        swapUINT32(&(fep->fe.RecLength));
        swapUINT64(&(fep->fe.InfoLengthH),&(fep->fe.InfoLengthL));
        swapUINT64(&(fep->fe.LogBlocksH),&(fep->fe.LogBlocksL));
        xlate_timestamp(&(fep->fe.sAccessTime));
        xlate_timestamp(&(fep->fe.sModifyTime));
        xlate_timestamp(&(fep->fe.sAttrTime));
        swapUINT32(&(fep->fe.Checkpoint));
        xlate_long_AD(&(fep->fe.sExtAttrICB));
        swapUINT64(&(fep->fe.UniqueIdH),&(fep->fe.UniqueIdL));
        xlate_implID( &(fep->fe.sImpID));
        /* sImpID - regid - no xlate */
        swapUINT32( &(fep->fe.L_EA));
        /* ADs */
        if (direction == FROM_OP) {
            swapUINT32(&(fep->fe.L_AD));
            xlate_ICBTag( &(fep->fe.sICBTag), direction);
        }
        maxBytes = MIN(maxBytes-FE_HdrLen(&(fep->fe)),fep->fe.L_AD);
        xlate_ADs(FE_ADStart(&(fep->fe)),
                  fep->fe.sICBTag.Flags & ADTYPEMASK,
                  maxBytes);
        if ( direction == TO_OP ) {
            swapUINT32(&(fep->fe.L_AD));
            xlate_ICBTag( &(fep->fe.sICBTag), direction);
        }
        break;
        
    case TAGID_INDIRECT:
        idep = (struct IndirectEntry *) dskstruct;
        xlate_tag( &(idep->sTag));
        xlate_ICBTag( &(idep->sICBTag), direction);
        xlate_long_AD( &(idep->sIndirectICB));
        break;
        
    case TAGID_ALLOC_EXTENT:
        AEDp = (struct AllocationExtentDesc *) dskstruct;
        xlate_tag(&AEDp->sTag);
        swapUINT32(&AEDp->prevAllocExtLoc);
        if ( direction == FROM_OP )
            swapUINT32(&AEDp->L_AD);
        maxBytes = MIN(maxBytes-AE_HdrLen,AEDp->L_AD);
        xlate_ADs(AE_ADStart(AEDp),ADType,
                  maxBytes);
        if ( direction == TO_OP )
            swapUINT32(&AEDp->L_AD);
        break;

    default:;
    }
} /* end ot_xlate */

/********************************************************************/
/*  ot_xlateBlock                                                   */
/*      translate a contiguous block of structured file entries     */
/*                                                                  */
/*  input parameters:                                               */
/*      entries:    buffer of structures to translate               */
/*      count:      the number to translate                         */
/*      entryType:  the type of structure                           */
/*      entrySize:  the size of each structure, in bytes            */
/*                                                                  */
/*  output parameters:                                              */
/*      entries:    contains the translated stuctures               */
/********************************************************************/
void ot_xlateBlock(caddr_t entries,
                   UINT16 count,
                   int entryType,
                   UINT16 entrySize)
{
    while (count-- > 0) {
        ot_xlate(entryType,TO_OP,entries,0,0);
        entries += entrySize;
    }
} /* end ot_xlateBlock */

/********************************************************************/
/*  getint                                                          */
/*      get a local 16-bit signed value from a buffer               */
/*                                                                  */
/*  input parameters:                                               */
/*      iptr:   buffer containing value                             */
/*                                                                  */
/*  returns:                                                        */
/*      the signed 16-bit value                                     */
/*                                                                  */
/*  notes:                                                          */
/*      assumes Intel format                                        */
/********************************************************************/
INT16 getint(caddr_t iptr)

{
    INT16   tmp;

    tmp = ((*(iptr+1) & 0xffL) << 8) | (*(iptr+0) & 0xffL);

    /* Sign-extend if high bit set */
    if (*(iptr+1) & 0x80)
        tmp |= ~0xffff;
    return tmp;
} /* end getint */

/********************************************************************/
/*  getuint                                                         */
/*      get a local unsigned 16-bit value from a buffer             */
/*                                                                  */
/*  input parameters:                                               */
/*      uptr:   buffer containing the value                         */
/*                                                                  */
/*  returns:    the unsigned 16-bit value                           */
/*                                                                  */
/*  notes:                                                          */
/*      assumes Intel format                                        */
/********************************************************************/
UINT16 getuint(caddr_t uptr)

{
    UINT16  tmp;

    tmp = ((*(uptr+1) & 0xffL) << 8) | (*(uptr+0) & 0xffL);
    return tmp;
} /* end getuint */

/********************************************************************/
/*  getlong                                                         */
/*      get a signed 32-bit value from a buffer                     */
/*                                                                  */
/*  input parameters:                                               */
/*      lptr:   buffer containing the value                         */
/*                                                                  */
/*  returns:                                                        */
/*      the signed 32-bit value                                     */
/*                                                                  */
/*  notes:                                                          */
/*      assumes Intel format                                        */
/********************************************************************/
INT32 getlong(register caddr_t lptr)

{
    INT32   tmp;

    tmp =   ((*(lptr+3) & 0xffL) << 24) |
            ((*(lptr+2) & 0xffL) << 16) |
            ((*(lptr+1) & 0xffL) <<  8) |
             (*(lptr+0) & 0xffL);

    /* Sign-extend if high bit set */
    if (*(lptr+3) & 0x80)
        tmp |= ~0xffffffffL;
    return tmp;
} /* end getlong */

/********************************************************************/
/*  getulong                                                        */
/*      get unsigned 32-bit value from a buffer                     */
/*                                                                  */
/*  input parameters:                                               */
/*      ulptr:  buffer containing value                             */
/*                                                                  */
/*  returns:                                                        */
/*      the unsigned 32-bit value                                   */
/*                                                                  */
/*  notes:                                                          */
/*      assumes Intel format                                        */
/********************************************************************/
UINT32 getulong(register caddr_t ulptr)

{
    UINT32  tmp;

    tmp =   ((*(ulptr+3) & 0xffL) << 24) |
            ((*(ulptr+2) & 0xffL) << 16) |
            ((*(ulptr+1) & 0xffL) <<  8) |
             (*(ulptr+0) & 0xffL);
    return tmp;
} /* end getulong */

/********************************************************************/
/*  putint                                                          */
/*      put a 16-bit value into a buffer                            */
/*                                                                  */
/*  input parameters:                                               */
/*      i:      the value                                           */
/*      iptr:   where to put it                                     */
/*                                                                  */
/*  output parameters:                                              */
/*      iptr:   contains the value in Intel format                  */
/*                                                                  */
/*  returns:                                                        */
/*      none                                                        */
/*                                                                  */
/********************************************************************/
void putint(UINT16 i,
            caddr_t iptr)

{
    *iptr = i & 0xffL;
    *(iptr+1) = i >> 8;
} /* end putint */

/********************************************************************/
/*  putlong                                                         */
/*      put a 32-bit value into a buffer                            */
/*                                                                  */
/*  input parameters:                                               */
/*      l:      the value                                           */
/*      lptr:   where to put it                                     */
/*                                                                  */
/*  output parameters:                                              */
/*      lptr:   contains the value in Intel format                  */
/*                                                                  */
/*  returns:                                                        */
/*      none                                                        */
/*                                                                  */
/********************************************************************/
void putlong(register UINT32 l,
             register caddr_t lptr)

{
    *(lptr+0) = l & 0xffL;
    *(lptr+1) = (l >>  8) & 0xffL;
    *(lptr+2) = (l >> 16) & 0xffL;
    *(lptr+3) = (l >> 24) & 0xffL;
} /* end putlong */

/********************************************************************/
/*  swapUINT16                                                      */
/*      swap bytes of a 16-bit value                                */
/*                                                                  */
/*  input parameters:                                               */
/*      iptr:   the value                                           */
/*                                                                  */
/*  output parameters:                                              */
/*      iptr:   the bytes have now been swapped                     */
/*                                                                  */
/*  returns:                                                        */
/*      none                                                        */
/*                                                                  */
/********************************************************************/
void swapUINT16(UINT16 *iptr)

{
    UINT16  tmp = *iptr;

    *((caddr_t) iptr+0) = tmp & 0xffL;
    *((caddr_t) iptr+1) = tmp >> 8;
} /* end swapUINT16 */

/********************************************************************/
/*  swapUINT32                                                      */
/*      swap bytes of a 32-bit value                                */
/*                                                                  */
/*  input parameters:                                               */
/*      lptr:   the value                                           */
/*                                                                  */
/*  output parameters:                                              */
/*      lptr:   the bytes have now been swapped                     */
/*                                                                  */
/*  returns:                                                        */
/*      none                                                        */
/*                                                                  */
/********************************************************************/
void swapUINT32(UINT32 *lptr)

{
    UINT32  tmp = *lptr;

    *((caddr_t) lptr+0) = tmp & 0xffL;
    *((caddr_t) lptr+1) = (tmp >>  8) & 0xffL;
    *((caddr_t) lptr+2) = (tmp >> 16) & 0xffL;
    *((caddr_t) lptr+3) = (tmp >> 24) & 0xffL;
} /* end swapUINT32 */

/*
 *  swapUINT64
 *      swap bytes of a 64-bit value
 *
 *  input parameters:
 *      hptr,lptr:   pointers to the high and low 32 bit locations.
 *
 *  output parameters:
 *      lptr:   the bytes have now been swapped
 *
 *  returns:
 *      none
 *
 */
void swapUINT64(UINT32 *hptr,UINT32 *lptr)

{
    UINT32  tmp;

    swapUINT32(hptr);
    swapUINT32(lptr);

    tmp   = *hptr;
    *hptr = *lptr;
    *lptr = tmp;
}

/********************************************************************/
/* xlate_tag - translate a ECMA 167 tag structure                   */
/*                                                                  */
/* Input:                                                           */
/*     tp - pointer to struct Tag to be translated                  */
/* Output:                                                          */
/*     tp - pointer to translated Tag                               */
/* Returns:                                                         */
/*     void                                                         */
/*                                                                  */
/* MODIFICATION HISTORY                                             */
/*  16Jul92  ECB  Created                                           */
/********************************************************************/
void xlate_tag( struct Tag *tp)
{
        swapUINT16( &(tp->uTagID));
        swapUINT16( &(tp->uDescriptorVersion));
        /* byte */
        /* byte */
        swapUINT16( &(tp->uTagSerialNum));
        swapUINT16( &(tp->uDescriptorCRC));
        swapUINT16( &(tp->uCRCLen));
        swapUINT32( &(tp->uTagLoc));
}
/********************************************************************/
/* xlate_ICBTag - translate an ECMA 167 ICBTag structure            */
/*                                                                  */
/* Input:                                                           */
/*     tp - pointer to struct Tag to be translated                  */
/* Output:                                                          */
/*     tp - pointer to translated Tag                               */
/* Returns:                                                         */
/*     void                                                         */
/*                                                                  */
/* MODIFICATION HISTORY                                             */
/*  19Nov92  ECB  Created                                           */
/********************************************************************/
void
xlate_ICBTag( struct ICBTag *tp, UINT32 direction)
{

    swapUINT32(&(tp->PriorDirects));

    /* how we transform the strategy parameter depends on the strategy type */

    if( direction == TO_OP) {
        if( tp->StrategyType == 1)
          swapUINT16((UINT16 *) &(tp->StrategyParm));
        swapUINT16(&(tp->StrategyType));
    } else {
        swapUINT16(&(tp->StrategyType));
        if( tp->StrategyType == 1)
          swapUINT16((UINT16 *)&(tp->StrategyParm));
    }

    swapUINT16(&tp->NumberEntries);
    /* byte reserved */
    /* byte FileType */
    swapUINT32(&tp->sParentICB_LBN);
    swapUINT16(&tp->sParentICB_PartNo);
    swapUINT16(&tp->Flags);
}

/********************************************************************/
/* xlate_domainID - translate an ISO13346 domain entity ID          */
/*                                                                  */
/* Input:                                                           */
/*     deip - pointer to domainEntityID to be translated            */
/* Output:                                                          */
/*     deip - pointer to translated Tag                             */
/* Returns:                                                         */
/*     void                                                         */
/*                                                                  */
/* MODIFICATION HISTORY                                             */
/*  09Apr96 DVM   Created                                           */
/********************************************************************/

void
xlate_domainID( struct domainEntityId *deip)
{
    /* uFlags       - no xlate */
    /* aID          - no xlate */
    swapUINT16(&(deip->uUDFRevision));
    /* uDomainFlags - no xlate */
    /* aReserved    - no xlate */
}
    
/********************************************************************/
/* xlate_udfID - translate an ISO13346 UDF-defined entity ID        */
/*                                                                  */
/* Input:                                                           */
/*     ueip - pointer to udfEntityID to be translated               */
/* Output:                                                          */
/*     ueip - pointer to translated Tag                             */
/* Returns:                                                         */
/*     void                                                         */
/*                                                                  */
/* MODIFICATION HISTORY                                             */
/*  09Apr96 DVM   Created                                           */
/********************************************************************/

void
xlate_udfID( struct udfEntityId *ueip)
{
    /* uFlags        - no xlate */
    /* aID           - no xlate */
    swapUINT16(&(ueip->uUDFRevision));
    /* uOSClass      - no xlate */
    /* uOSIdentifier - no xlate */
    /* aReserved     - no xlate */
}
    
/********************************************************************/
/* xlate_implID - translate ISO13346 implementation-use entity ID   */
/*                                                                  */
/* Input:                                                           */
/*     ieip - pointer to implEntityID to be translated              */
/* Output:                                                          */
/*     ieip - pointer to translated Tag                             */
/* Returns:                                                         */
/*     void                                                         */
/*                                                                  */
/* MODIFICATION HISTORY                                             */
/*  09Apr96 DVM   Created                                           */
/********************************************************************/

void
xlate_implID( struct implEntityId *ieip)
{
    /* uFlags        - no xlate */
    /* aID           - no xlate */
    /* uOSClass      - no xlate */
    /* uOSIdentifier - no xlate */
    /* uDescVersion  - no xlate */
    /* uLibrVersion  - no xlate */
    swapUINT16(&(ieip->uImplRegNo));
    /* uImplVersion  - no xlate */
    /* uImplRevision - no xlate */
}
    
/********************************************************************/
/* xlate_extent - translate a ECMA 167 extent_ad structure          */
/*                                                                  */
/* Input:                                                           */
/*     exp - pointer to struct Tag to be translated                 */
/* Output:                                                          */
/*     exp - pointer to translated Tag                              */
/* Returns:                                                         */
/*     void                                                         */
/*                                                                  */
/* MODIFICATION HISTORY                                             */
/*  16Jul92  ECB  Created                                           */
/********************************************************************/

void
xlate_extent( struct extent_ad *exp)
{
        swapUINT32(&exp->Length);
        swapUINT32(&exp->Location);
}


/********************************************************************/
/* xlate_short_AD - translate a ECMA 167 short_AD structure         */
/*                                                                  */
/* Input:                                                           */
/*     exp - pointer to struct short_AD to be translated            */
/* Output:                                                          */
/*     exp - pointer to translated structure                        */
/* Returns:                                                         */
/*     void                                                         */
/*                                                                  */
/* MODIFICATION HISTORY                                             */
/*  092492  ECB  Created                                            */
/********************************************************************/

void
xlate_short_AD( struct short_AD *lap)
{
    swapUINT32(&lap->ExtentLength.Length32);
    swapUINT32(&lap->Location);
}

/********************************************************************/
/* xlate_long_AD - translate a ECMA 167 long_AD structure           */
/*                                                                  */
/* Input:                                                           */
/*     exp - pointer to struct long_AD to be translated             */
/* Output:                                                          */
/*     exp - pointer to translated structure                        */
/* Returns:                                                         */
/*     void                                                         */
/*                                                                  */
/* MODIFICATION HISTORY                                             */
/*  092492  ECB  Created                                            */
/********************************************************************/

void
xlate_long_AD( struct long_AD *lap)
{
    swapUINT32(&lap->ExtentLength.Length32);
    swapUINT32(&lap->Location_LBN);
    swapUINT16(&lap->Location_PartNo);
    swapUINT16(&lap->uUdfFlags);
    /* impuse - no xlate */
}

/********************************************************************/
/* xlate_ext_AD - translate a ECMA 167 ext_AD structure             */
/*                                                                  */
/* Input:                                                           */
/*     eap - pointer to struct ext_AD to be translated              */
/* Output:                                                          */
/*     eap - pointer to translated structure                        */
/* Returns:                                                         */
/*     void                                                         */
/*                                                                  */
/* MODIFICATION HISTORY                                             */
/*  092492  ECB  Created                                            */
/********************************************************************/

void
xlate_ext_AD(struct ext_AD *eap)
{
    swapUINT32(&eap->ExtentLength.Length32); /* MAY want to explicitly */
                                             /* swap bitfields here. */
    swapUINT32(&eap->RecordedLength);
    swapUINT32(&eap->InfoLength);
    swapUINT32(&eap->Location_LBN);
    swapUINT16(&eap->Location_PartNo);
    /* impuse - no xlate */
}

/*
 * xlate_ADs()
 *
 * This does the translation of all the ADs starting at the supplied
 * address (ADp) going for the supplied number of bytes (L_AD), based on
 * the supplied type of AD (ADType).
 *
 */

void
xlate_ADs(void *ADp,
               UINT32 ADType,
               UINT32 L_AD)
{
    UINT32 i;
    UINT32 numADs = L_AD/ot_sizeOfAD(ADType);

    switch (ADType) {
    case ADSHORT: 
        for (i = 0L; i < numADs; i++)
            xlate_short_AD(((struct short_AD *)ADp)+i);
        break;
    case ADLONG: 
        for (i = 0L; i < numADs; i++)
            xlate_long_AD(((struct long_AD *)ADp)+i);
        break;
    case ADEXTENDED:
        for (i = 0L; i < numADs; i++)
            xlate_ext_AD(((struct ext_AD *)ADp)+i);
        break;
    case ADNONE:
        numADs = 1; /* file contained within this file entry */
        break;
    default:
        DPRINTF("xlate_ADs: unknown AD type.\n");
        break;
    }
}

/********************************************************************/
/* xlate_timestamp - translate a ECMA 167 timestamp structure       */
/*                                                                  */
/* Input:                                                           */
/*     tsp - pointer to struct timestamp to be translated           */
/* Output:                                                          */
/*     tsp - pointer to translated structure                        */
/* Returns:                                                         */
/*     void                                                         */
/*                                                                  */
/* MODIFICATION HISTORY                                             */
/*  23Nov92  ECB  Created                                           */
/********************************************************************/
void
xlate_timestamp( struct timestamp *tsp)

{
    /* only two things need swapping in a timestamp */
    swapUINT16((UINT16 *) &(tsp->uTypeAndTimeZone));
    swapUINT16((UINT16 *) &(tsp->iYear));
}

/********************************************************************/
/* xlate_fid - translate a ECMA 167 file identifier                 */
/*                                                                  */
/* Input:                                                           */
/*     fidp - pointer to struct FileIDDesc                          */
/* Output:                                                          */
/*     fidp - pointer to translated structure                       */
/*                                                                  */
/* MODIFICATION HISTORY                                             */
/********************************************************************/
void
xlate_fid(struct FileIDDesc *fidp,INT32 nbytes,UINT32 direction)
{
    struct Tag *tp = (struct Tag *)fidp;

    if ( nbytes >= FILE_ID_DESC_CONSTANT_LEN ) {
        xlate_tag( &(fidp->sTag));
        swapUINT16( &(fidp->VersionNum));
        /* UCHAR8 Characteristics -- no xlate */
        /* UCHAR8 L_FI -- no xlate */
        xlate_long_AD( &(fidp->ICB));
        if( direction == FROM_OP)  {
            swapUINT16( &(fidp->L_IU));
        }
        nbytes -= FILE_ID_DESC_CONSTANT_LEN;

        if ( nbytes >= fidp->L_IU ) {
            if ( fidp->L_IU >= sizeof(struct implEntityId) ){
                xlate_implID(& (fidp->sImplementationID));
            }
            /* Implementation use, no xlate */
            nbytes -= fidp->L_IU;
            if (nbytes >= fidp->L_FI ) {
                /* file identifier, no xlate */
                nbytes -= fidp->L_FI;
            }
        }
        if ( direction == TO_OP ) {
            swapUINT16( &(fidp->L_IU));
        }
    }
}
#endif /* not IBM */

/********************************************************************/
/*  ot_copy                                                         */
/*      copy bytes from one string to another                       */
/*                                                                  */
/*  input parameters:                                               */
/*      s:  source string                                           */
/*      d:  destination string                                      */
/*      n:  number of bytes to copy                                 */
/********************************************************************/
void
ot_copy(register caddr_t s,
             register caddr_t d,
             register INT32 n)

{
    memmove((void *)d,(void *)s, (size_t)n);
} /* end ot_copy */

/******************************************************************/
/*                                                                */
/*  Don't support the creation of the TAG CRC field               */
/*                                                                */
/******************************************************************/
void
FillTag( UINT16 DescID,
         UINT16 DescVersion,
         UINT16 TagSerial,
         UINT16 CalcCRC,
         UINT16 CRCLen,
         UINT32 Location,
         struct Tag *TagPtr)
{
    UCHAR8 *BytePtr;
    UCHAR8 Checksum = 0;
    UINT16 i;
    
    BytePtr = (UCHAR8 *) TagPtr;
    
    TagPtr->uTagID = DescID;
    TagPtr->uDescriptorVersion = DescVersion;
    TagPtr->uTagChecksum = 0;
    TagPtr->uTagSerialNum = TagSerial;
    TagPtr->uReserved = 0;

    /* we don't support Descriptor CRC's right now */
    TagPtr->uDescriptorCRC = 0;
    TagPtr->uCRCLen = 0;
    
    TagPtr->uTagLoc = Location;

    for (i=0; i<16; i++, BytePtr++)
        Checksum+= *BytePtr;
    TagPtr->uTagChecksum = Checksum;
}

/******************************************************************/
/*                                                                */
/*  Given a pointer to a suspected Tag, return the tag value if   */
/*  it is a tag, else return NOTAG                                */
/*                                                                */
/*                                                                */
/******************************************************************/
UINT16
isTag( struct Tag *TagPtr)

{
    UCHAR8 *BytePtr;
    UCHAR8 Checksum = 0;
    UINT16 i;

    BytePtr = (UCHAR8 *) TagPtr;

    for (i=0; i<16; i++, BytePtr++)
        Checksum+= *BytePtr;

    if (Checksum == TagPtr->uTagChecksum)
      return TagPtr->uTagID;
    else
      return TAGID_NONE;
}

/*
 * ot_goodTag()
 *
 * Return Success if the supplied tag's id  matches the supplied
 * tagid, the checksum is OK, and the Location field is correct.
 *
 */
UINT16
ot_goodTag(UINT16 expected_tagid,
           void *ptr,
           UINT32 Location)
{
    struct Tag *tagp = ptr;
    UINT16 i;
    UCHAR8 CheckSum;
    UCHAR8 *BytePtr;

    /* check to see if a valid tag, start by checking location field
       and then see if checksum is OK */

    if( tagp->uTagLoc != Location)
        return EINVAL;

    /* test the tagid value */
    if ( expected_tagid != TAGID_NONE ) {
        /* go ahead and check to see if tagid's match */
        if ( expected_tagid != tagp->uTagID ) {
            return EINVAL;
        }
    }

    /* test the checksum on the descriptor */
    BytePtr = (UCHAR8 *) tagp;
    CheckSum = 0;
      
    for (i=1; i<5; i++, BytePtr++)
        CheckSum+= *BytePtr;
    *BytePtr++;                            /* skip the 8 bit checksum field */
    for (i=1; i<12; i++, BytePtr++)
        CheckSum+= *BytePtr;
            
    if( tagp->uTagChecksum != CheckSum)
        return EINVAL;
    
    return(SUCCESS);
}

#ifdef DO_CRC
/* 
* CRC for x^16 + x^12 + x^5 + 1
* octal 0210041 = hex 0x11021 = decimal 69665 
*/
static unsigned short crc_table[256]={
     0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7, 
     0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef, 
     0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6, 
     0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de, 
     0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485, 
     0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d, 
     0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4, 
     0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc, 
     0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823, 
     0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b, 
     0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12, 
     0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a, 
     0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41, 
     0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49, 
     0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70, 
     0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78, 
     0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f, 
     0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067, 
     0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e, 
     0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256, 
     0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d, 
     0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405, 
     0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c, 
     0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634, 
     0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab, 
     0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3, 
     0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a, 
     0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92, 
     0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9, 
     0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1, 
     0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8, 
     0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1EF0
};

static unsigned short cksum(register unsigned char *s,register int n)
{
    register unsigned short crc=0;
    
    while (n-- > 0)
        crc = crc_table[(crc>>8 ^ *s++) &0xff] ^ (crc<<8);

    return crc;
}

#ifdef CRC_EXAMPLE
unsigned char bytes[] = { 0x70, 0x6A, 0x77 };

main ()
{
     unsigned short x;

     x = cksum(bytes, sizeof bytes);
     printf("checksum: calculated=%4.4x, correct=%4.4x\n", x, 0x3299);
     exit (0);
}
#endif /* CRC_EXAMPLE */


/* The crc_table[] was generated by the following program: */

#ifdef GENERATE_CRC_TABLE
/*
 *  a.out 0x11021 for CRC-CCITT
 */
main(int argc,char *argv[])
{
    unsigned long crc, poly;
    int n, i;

    sscanf(argv[1],"%lo", &poly);
    printf("CRC 0%o \n", poly);
    printf("static unsigned short crc_table[256]=(\n");
    for (n = 0; n < 256; n++) 
    {
         if (n % 8 == 0)
             printf("     ");
         crc = n <<8;
         for(i = 0; i<8; i++)
         {
             if(crc & 0x8000)
                 crc = (crc<<1)^poly;
             else
                 crc <<= 1;
             crc &= 0xFFFF;
         }
         if (n == 255)
              printf("0x%04X", crc);
         else
              printf("0x%04x, ", crc);
         if (n%8==7)
              printf("\n");
   }
   printf(");\n");
   exit (0);
}
#endif /* GENERATE_CRC_TABLE */
#endif /* DO_CRC */
