/********************************************************************/
/*  o_vfsops.c - NSR routines for vfs/otfs operations               */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/*                                                                  */
/*  Module name:    o_vfsops.c                                      */
/*                                                                  */
/*  Description:    VFS routines for the management of file         */
/*                  systems; routines to manage otnode lists;       */
/*                  routines to manage disk space                   */
/*                                                                  */
/*  Functions in this module:                                       */
/*      otfs_root           otfs_statfs                             */
/*      otfs_sync           otfs_vget                               */
/*                                                                  */
/*      ot_active           ot_addToList        ot_busy             */
/*      ot_calcSpaceToFlush ot_dequeueOTNode    ot_enqueueOTNode    */
/*      ot_eraseUnerased     ot_getContiguous    ot_getSpace        */
/*      ot_putOnline        ot_rdHighwaterMark  ot_rdwrOBase        */
/*      ot_readBitmap       ot_takeFromList     ot_writeBitmap      */
/*      ot_wtHighwaterMark                                          */
/*                                                                  */
/*      killBitmap          killSpaceList       power               */
/*      returnSectors       uioInit             vattrInit           */
/*                                                                  */
/*  Modification history:                                           */
/*      18.Mar.92   ECB     Added ANSI prototypes                   */
/*      22.Sep.92   ECB     Change ot_countfreesectors and removed  */
/*                          isHPport()                              */
/*                                                                  */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrCore/o_vfsops.c,v $
 $Revision: 1.37 $      $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:53:02 $
 ********************************************************************/

#include "Debug.h"
#include "../nsrHdrs/def.h"
#include "../nsrHdrs/o_vfsops.h"
#include "../nsrHdrs/nsr.h"
#include "../nsrHdrs/o_proto.h"
#include "../nsrHdrs/o_node.h"
#include "../nsrHdrs/nsrerrno.h"

#if (MACINTOSH || IBM || HP)
struct vfsops otfs_vfsops;
#endif

#if SUN
struct vfsops otfs_vfsops = {
    otfs_mount_lck,     /* vfs_mount    */
    otfs_unmount,       /* vfs_unmount  */
    otfs_root_lck,      /* vfs_root     */
    otfs_statfs_lck,    /* vfs_statfs   */
    otfs_sync_lck,      /* vfs_sync     */
    otfs_vget_lck,      /* vfs_vget     */
};
#endif /* SUN */

#if HP
struct vfsops otfs_vfsops = {
    (int (*)()) otfs_mount_dcp, /* vfs_mount     */
    (int (*)()) otfs_unmount,   /* vfs_unmount   */
    (int (*)()) otfs_root_dcp,  /* vfs_root      */
    (int (*)()) otfs_statfs,    /* vfs_statfs    */
    (int (*)()) otfs_sync,      /* vfs_sync      */
    (int (*)()) otfs_vget,      /* vfs_vget      */
    (int (*)()) otfs_invalop,   /* vfs_mountroot */
};
#endif /* HP */

/* globals */
static UCHAR8   bitPositions[8] = { BITZERO, BITONE, BITTWO, BITTHREE, BITFOUR, BITFIVE,
                                    BITSIX, BITSEVEN };
extern  INT32   otfs_debug;

/********************************************************************/
/*  otfs_invalop:                                                   */
/*      VFS routine to indicate an invalid request for this file    */
/*      system was made.                                            */
/*                                                                  */
/*  returns:                                                        */
/*      EINVAL: inavlid request                                     */  
/********************************************************************/
INT32 otfs_invalop()
{
    return EINVAL;
} /* end otfs_invalop */

/********************************************************************/
/*  otfs_root                                                       */
/*      VFS routine to get the vnode for the root of the volume     */
/*                                                                  */
/*  input parameters:                                               */
/*      vfsp:       the vfs struture for the volume                 */
/*      vpp:        for the root vnode                              */
/*      version:    the version of the root to get                  */
/*                                                                  */
/*  output parameters:                                              */
/*      vpp:        contains version of the root desired, if found  */
/*                                                                  */
/*  returns:                                                        */
/*      0           success                                         */
/*      ENOMEM:     memory allocation failure                       */
/********************************************************************/
INT32 otfs_root(struct vfs *vfsp,
                struct vnode **vpp,
                UINT32 version)

{
    INT32                   error; 
    struct otnode           *otp;
    register struct otfs    *fsp = (struct otfs *) vfsp->vfs_data;

    /* get an otnode to use */
    DMLPRINTF(oVfsLev)("otfs_root: rootAddr= %lx, version = %lx, volName = %s\n",
        fsp->rootAddr, version, fsp->VolumeName);

    error = ot_getOTNode( 0, fsp, fsp->VolumeName, fsp->rootAddr, version, VDIR,
                        (struct otnode *) 0, &otp);
    if (!error) {
        otp->vnode.v_flag = VROOT;
        *vpp = &otp->vnode;
    }
    return error;
} /* end otfs_root */

/********************************************************************/
/*  otfs_statfs                                                     */
/*      VFS routine to get file system information                  */
/*                                                                  */
/*  input parameters:                                               */
/*      vfsp:   the vfs structure for the volume                    */
/*      sbp:    for the status information                          */
/*                                                                  */
/*  output parameters:                                              */
/*      sbp:    sbp->f_bsize = sector size                          */
/*              sbp->f_bfree = # sectors free on the volume         */
/*              sbp->f_blocks = size of volume                      */
/*              sbp->f_bavail = # sectors free on the volume        */
/*  returns:                                                        */
/*      0       success                                             */
/********************************************************************/
INT32 otfs_statfs(struct vfs *vfsp, 
                  register struct statfs *sbp)

{
    register struct otfs *fsp = (struct otfs *) vfsp->vfs_data;
    bzero((caddr_t)sbp,sizeof(struct statfs));
    
    sbp->f_type     = (INT32) 0;
    sbp->f_bsize    = (INT32) fsp->SectorSize;
    sbp->f_blocks   = fsp->VolSection[0].volsiz;
    sbp->f_bfree    = (INT32) fsp->sectorsFree;
    sbp->f_bavail   = (INT32) fsp->sectorsFree;
    sbp->f_files    = (INT32) 0;
    sbp->f_ffree    = (INT32) 0;
    sbp->f_fsid     = vfsp->vfs_fsid;
    sbp->f_spare[0] = 0;        /* For Later? */
    sbp->f_spare[1] = 0;        /* For Later? */
    return 0;
} /* end otfs_statfs */

/********************************************************************/
/*  otfs_sync                                                       */
/*      VFS routine to flush a volume; guarantees consistency of    */
/*      the file system at the end of a flush request               */
/*                                                                  */
/*  input parameters:                                               */
/*      vfsp:   the vfs structure for the volume                    */
/*                                                                  */
/*  returns:                                                        */
/*      0       success                                             */
/*                                                                  */
/********************************************************************/
INT32 otfs_sync(struct vfs *vfsp)

{
    struct otnode           *next;
    struct bitmapInfo       *info;
    register INT32          error;
    register struct otnode  *active;
    register struct otfs    *fs = (struct otfs *) vfsp->vfs_data;

    error = 0;

    /* flush all the modified icb entries except for the erased bitmap, we take
        care of that separately */
    for (active = fs->active; active && !error; active = next) {
        next = active->next;
        if (active->icbe.ge.sICBTag.sParentICB_LBN != fs->erasedAddr)
            error = otvn_fsync(&active->vnode,(struct ucred *) 0);
    }

    if (fs->MediaType <= (int) fsWORM) {    /* flush highwater mark */
        if (!error)
            error = ot_wtHighwaterMark(fs);
    } else {                                /* flush bitmaps */
        if (!error && fs->unerasedBitmap)
            error = ot_eraseUnerased(fs);
        if (!error && fs->erasedBitmap)
            if (fs->erasedBitmap->bitmapData && (fs->erasedBitmap->modified & OT_MOD))
                error = ot_writeBitmap(fs,bmapErased);
        if (!error && fs->unerasedBitmap)
            if (fs->unerasedBitmap->bitmapData && (fs->unerasedBitmap->modified & OT_MOD))
                error = ot_writeBitmap(fs,bmapUnerased);
    }

    /* Then, tell the buffer cache to flush. */
    if ( !error ) {
        ot_bflush(fs);
    }
    return error;
} /* end otfs_sync */

/********************************************************************/
/*  otfs_vget                                                       */
/*      VFS routine to turn a unique identifier into a vnode        */
/*                                                                  */
/*  input parameters:                                               */
/*      vfsp:       the vfs structure for the volume                */
/*      vpp:        for the new vnode                               */
/*      fidp;       contains the unique file identification info    */
/*                                                                  */
/*  output parameters:                                              */
/*      vpp:        newly allocated, and contains the vnode for the */
/*                  file/directory                                  */
/*  returns:                                                        */
/*      0:          success                                         */
/*      ENOMEM:     a memory allocation failed                      */
/*                                                                  */
/********************************************************************/
INT32 otfs_vget(struct vfs *vfsp,
                struct vnode **vpp,
                struct fid *fidp)

{
    struct otnode           *otp;
    struct otfs             *fsp = (struct otfs *) vfsp->vfs_data;
    register INT32          error;
    register struct fidData *fidData;

    fidData = (struct fidData *) fidp->fid_data;

    DMLPRINTF(oVfsLev)("otfs_vget: top\n");

    /* get an otnode to use */
    error = ot_getOTNode( 0, fsp, fidData->name, fidData->icbAddr, fidData->version,
                        fidData->attribute, (struct otnode *) 0, &otp);
    if (!error) {
        *vpp = &otp->vnode;

        /*  Note that Sun Versions through 3.X do not want the structure freed, but every other */
        /*  implementation does. */

        if (ORIG_ID != SUN_ORIG_ID)
            kmem_free((caddr_t) fidp,(UINT32) sizeof(struct fid) - MAXFIDSZ + fidp->fid_len,0);
    }

    DMLPRINTF(oVfsLev)("otfs_vget: bottom\n");

    return error;
} /* end otfs_vget */

/********************************************************************/
/*  ot_active                                                       */
/*      routine to determine if an otnode is active                 */
/*                                                                  */
/*  input parameters:                                               */
/*      otp:    the otnode to check                                 */
/*                                                                  */
/*  returns:                                                        */
/*      0       not active                                          */
/*      1       active                                              */
/*                                                                  */
/*  notes:                                                          */
/*      an otnode is active if it is being referenced, or it has    */
/*          been modified                                           */
/********************************************************************/
INT32 ot_active(struct otnode *otp)

{
    DMLPRINTF(oVfsLev)("ot_active: top\n");

    if ((otp->vnode.v_count != 0 || (otp->flags & OT_MOD)) &&
        !(otp->flags & OT_INVAL)) {
        DMLPRINTF(oVfsLev)("ot_active: bottom\n");
        return 1;
    } else  {
        DMLPRINTF(oVfsLev)("ot_active: bottom\n");
        return 0;
    }
} /* end ot_active */

/********************************************************************/
/*  ot_addToList                                                    */
/*      routine to add space information to a list                  */
/*                                                                  */
/*  input parameters:                                               */
/*      start:      starting address of information to add          */
/*      length:     number of sectors in this chunk                 */
/*      listptr:    the list to add it to                           */
/*      fsp:        the otfs structure for the volume               */
/*                                                                  */
/*  output parameters:                                              */
/*      listptr:    may or may not have changed                     */
/*                                                                  */
/*  returns:                                                        */
/*      0           success                                         */
/*      ENOMEM:     memory allocation failure                       */
/********************************************************************/
INT32 ot_addToList(register UINT32 start,
                   register UINT32 length,
                   struct space **listptr,
                   struct otfs *fs)
{
    register struct space   *mark,
                            *ptr,
                            *tmp;
    INT32                   error;
    UINT32                  oldLength = length;

    error = 0;

    /*
     * if it's WORM and the erased space list does exist, don't add
     * to it. if it doesn't exist, it's ok to create it. The theory is
     * that a WORM erased list should always be one link long and that the
     * recovery of WORM space is meaningless.
     */
    if (fs->MediaType <= (int) fsWORM && listptr == &fs->highWater && *listptr)
        return 0;

    for (ptr = *listptr, mark = ptr;
            ptr && ptr->start+ptr->length < start;
            mark = ptr, ptr = ptr->next);
    if (ptr && ptr->start <= start+length) {     /* then merge into ptr */
        if (start < ptr->start) {
            ptr->length += ptr->start - start;
            ptr->start = start;
        } else {
            length += start - ptr->start;
            if (ptr->length < length)
                ptr->length = length;
            mark = ptr->next;
            if (mark && mark->start <= ptr->start+ptr->length) {
                ptr->length = (mark->start+mark->length) - ptr->start;
                ptr->next = mark->next;
                if (mark->next)
                    mark->next->previous = ptr;
                kmem_free((char *) mark,(UINT32) sizeof(struct space),0);
            }
        }
    } else {                                        /* then add onto mark */
        /* (after mark, not before ptr, because ptr may be at list's end (0)) */
        tmp = ptr;
        ptr = (struct space *) kmem_alloc((UINT32) sizeof(struct space),0);
        if (ptr) {
            ptr->modified = (UINT16) 0;
            ptr->start  = start;
            ptr->length = length;
            if (tmp == *listptr) {
                *listptr = ptr;
                ptr->next = mark;
                ptr->previous = 0;
                if (mark)   /* will be 0 if list was entirely unallocated */
                    mark->previous = ptr;
            } else {
                ptr->next = mark->next;
                ptr->previous = mark;
                mark->next = ptr;
                if (ptr->next)
                    ptr->next->previous = ptr;
            }
        } else
            error = ENOMEM;
    }
    if (!error && listptr == &fs->highWater)
        fs->sectorsFree += oldLength;
    return error;
} /* end ot_addToList */

/********************************************************************/
/* ot_busy() -  routine to check if the file system is busy         */
/********************************************************************/
INT32 ot_busy(struct otfs *fsp)

{
    return 0;
} /* end ot_busy */

/********************************************************************/
/*  ot_calcSpaceToFlush                                             */
/********************************************************************/
UINT32 ot_calcSpaceToFlush(struct otfs *fs,
                           UINT32 version)

{
    UINT16      level,          /* number of indirects counter */
                offset;         /* within a block */
    UINT32      levelCount,     /* # versions at a level of indirection*/
                vsnCount,       /* version counter as we go along */
                spRequired,     /* # sectors required to flush this version */
                ctr, ctr2;      /* loop counters */


    /* find out what level of indirection icb entry will be at */

    for (levelCount=1,vsnCount=fs->ICBsize-fs->IndirectSize,level = 0;
         vsnCount <= version && level < fs->IndirectSize;
         level++, levelCount *= fs->ICBsize, vsnCount += levelCount);
    
    if (level == 0)
      offset = version;
    else {
        offset = fs->ICBsize - fs->IndirectSize + level - 1;
        version -= vsnCount - levelCount;
        while (level > 0 && version != 0) {
            levelCount /= fs->ICBsize;
            offset = version/levelCount;
            version %= levelCount;
            level--;
        }
    }
    /* do we need to allocate a new direct block? */
    spRequired = offset == 0 ? fs->ICBsize : 0;
    
    /* add in the number of indirect blocks we'd have to allocate... */
    spRequired += level * fs->ICBsize;

    return spRequired;
} /* end ot_calcSpaceToFlush */

/********************************************************************/
/*  ot_countFreeSectors                                             */
/********************************************************************/
INT32 ot_countFreeSectors(register struct otfs *fs)

{
    INT32               error;
    UCHAR8              bitpos;
    UINT32              idx;
    UINT32              bit;
    UCHAR8              *addr;       /* ptr into bitmap */
    UCHAR8              *addr_last;
    INT32               bsize;
    UINT32              nbits;
    INT32               sector;
    struct SpaceBitmapEntry *sb; 

    
    /* allocate a buffer to read in the bitmap data */
    sb = (struct SpaceBitmapEntry *) kmem_alloc((UINT32)fs->SectorSize,BUFRAM);

    /* read in the space bitmap entry */
    sector = fs->erasedAddr;
    bsize = fs->SectorSize;
    if ( error = ot_read(fs,sector++,(caddr_t)sb,&bsize) ) {
        DPRINTF("ot_countFreeSectors: can't read erased bitmap\n");
        return error;
    }
#if !IBM
    if ( ! error ) 
        ot_xlate(TAGID_SPACE_BMAP,FROM_OP,(caddr_t)sb,0,0);
#endif
    
    nbits = sb->hdr.N_Bits;
    
    fs->sectorsFree = 0;

    addr = &sb->bitmap[0];
    addr_last = (UCHAR8 *)sb + fs->SectorSize;
    for ( bit = 0, idx = 0; bit < nbits && !error ; bit++ )
    {
        bitpos = bit % 8;
        if ( *addr & bitPositions[bitpos] ) {
            fs->sectorsFree++;
        }
        if ( bitpos == 7 ) {
            addr++;
            if ( addr >= addr_last ) {
                if ( error = ot_read(fs,sector++,(caddr_t)sb,&bsize) ) {
                    DPRINTF("ot_countFreeSectors: can't read erased bitmap\n");
                    return error;
                }
                addr = (UCHAR8 *)sb;
            }
        }
    }
    
    DMLPRINTF(oVfsLev)("ot_countFreeSectors: Used bits in bitmap for sectorsFree=%lu.\n",
                       fs->sectorsFree);
    
    /* free the buffer allocated earlier */
    kmem_free((caddr_t)sb,fs->SectorSize,BUFRAM);
    
    return error;
} /* end ot_countFreeSectors */


/********************************************************************/
/*  ot_dequeueOTNode                                                */
/*      routine to delete an otnode from an otnode list             */
/*                                                                  */
/*  input parameters:                                               */
/*      listptr:    head of the list                                */
/*      otp:        the one to remove from the list                 */
/*                                                                  */
/*  output parameters:                                              */
/*      listptr:    may or may not have changed                     */
/*                                                                  */
/*  returns:                                                        */
/*      none                                                        */
/*                                                                  */
/********************************************************************/
void ot_dequeueOTNode(struct otnode **listptr,
                      register struct otnode *otp)

{
    DMLPRINTF(oVfsLev)("dequeueOTNode: removing otnode at %lx from %p\n",
        otp->icb_addr_LBN,(void *)*listptr);

    if (otp->next)
        otp->next->previous = otp->previous;
    if (otp->previous)
        otp->previous->next = otp->next;
    else
        *listptr = otp->next;
    otp->next = (struct otnode *) 0;
    otp->previous = (struct otnode *) 0;

/*  print_fs_ICBlists((struct otfs *) otp->vnode.v_vfsp->vfs_data); */
} /* end ot_dequeueOTNode */

/********************************************************************/
/*  ot_enqueueOTNode                                                */
/*      insert an otnode into an otnode list                        */
/*                                                                  */
/*  input parameters:                                               */
/*      listptr:    head of the list                                */
/*      otp:        the one to add                                  */
/*                                                                  */
/*  output parameters:                                              */
/*      listptr:    otp is now the head of the list                 */
/*                                                                  */
/*  returns:                                                        */
/*      none                                                        */
/*                                                                  */
/********************************************************************/
void ot_enqueueOTNode(struct otnode **listptr,
                      register struct otnode *otp)

{
    DMLPRINTF(oVfsLev)("enqueueOTNode: adding otnode at %lx on %p\n",
        otp->icb_addr_LBN,(void *)*listptr);

    /* insert new otnode at head of list */
    otp->next = *listptr;
    otp->previous = (struct otnode *) 0;
    if (otp->next)
        otp->next->previous = otp;
    *listptr = otp;

/*  print_fs_ICBlists((struct otfs *) otp->vnode.v_vfsp->vfs_data); */
} /* end ot_enqueueOTNode */

#define OFFSET(type, field)     ((int) &((type *) 0)->field)
/********************************************************************/
/*  ot_eraseUnerased                                                */
/*      erase sectors found in the cached portion of the unerased   */
/*      sector bitmap. NOTE that this currently erases ALL sectors  */
/*      in the unerased bitmap.                                     */
/*                                                                  */
/*  input parameters:                                               */
/*      fs: the otfs structure for the volume                       */
/*                                                                  */
/*  returns:                                                        */
/*      0       success                                             */
/********************************************************************/
INT32 ot_eraseUnerased(register struct otfs *fs)

{
    UINT32              perSector,  /* # bytes per sector used for bitmap data */
                        byteInFile, /* byte offset into bitmap, not counting headers */
                        address;    /* calculated volume block address */
    struct bitmap       *bmap;
    register int        idx;
    register INT32      error;
    register caddr_t    data,
                        dataEnd;

    if (fs->MediaType <= (int) fsWORM)
        return 0;

    error = 0;
    perSector = fs->SectorSize;
    bmap = fs->unerasedBitmap;
    data = bmap->bitmapData;
    dataEnd = bmap->bitmapEnd;

    /* ISSUE: this needs to be optimized to scan for series of sectors */

    /*
    * look at each bit of each byte of the cached unerased bitmap sector. If set,
    * erase the sector.  if erasure successful, reset bit and return the sector
    * to the erased bitmap 
    */ 
    for (; data <= dataEnd && !error; data++) {
        for (idx = 0; idx < bitsPerByte && !error; idx++) {
            if (*data & bitPositions[idx]) {    /* bit set */
                byteInFile = bmap->logSector * perSector +
                    (data-(bmap->bitmapData));
                address = byteInFile * bitsPerByte + idx;

                /*  all space in unerased bitmap list is deleted file data, which we don't bother
                    actually erasing (thus the 0 parameter) unless necessary ?????*/
                error = ot_erase(fs,(int) 0,address,(UINT32) 1);
                if (!error) {
                    *data &= ~bitPositions[idx];
                    fs->unerasedBitmap->modified |= OT_MOD;
                    error = returnSectors(fs,(struct space *) 0,address,(UINT32) 1,bmapErased);
                }
            }
        }
    }
    return error;
} /* end ot_eraseUnerased */

/********************************************************************/
/*  ot_getContiguous                                                */
/*      routine to get contiguous space on the disk to use          */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:    the otfs structure for the volume                   */
/*      size:   the number of sectors needed                        */
/*                                                                  */
/*  returns:                                                        */
/*      0                   no space on device or memory allocation */
/*                              failure                             */
/*      pointer to link     success                                 */
/********************************************************************/
struct space *ot_getContiguous(register struct otfs *fs,
                               UINT32 size)

{
    int                     firstBit;   /* first bit in a chain? */
    INT32                   idx;
    UINT32                  byteOffset, /* byte offset into bitmap data, not counting headers */
                            perSector;  /* number of bytes per sector used for bitmap data */
    caddr_t                 data,
                            dataEnd;
    struct bitmap           *bmap;
    register INT32          error;
    register UINT32         reqSize;    /* decrementing counter */
    register struct space   *space;

    if (fs->MediaType <= (int) fsWORM) {
        space = (struct space *) 0;
        if (fs->highWater->length >= size) {
            space = (struct space *) kmem_alloc((UINT32) sizeof(struct space),0);
            if (space) {
                space->modified   = (UINT16) 0;
                space->next = (struct space *) 0;
                space->previous = (struct space *) 0;
                space->start = fs->highWater->start;
                space->length = size;
                fs->highWater->start += size;
                fs->highWater->length -= size;
                fs->highWater->modified |= OT_MOD;
                fs->sectorsFree -= size;
            }
        }
        return space;
    }

    error = 0;
    space = (struct space *) kmem_alloc((UINT32) sizeof(struct space),0);
    if (!space)
        error = ENOMEM;
    else
        bzero((caddr_t) space,(INT32) sizeof(struct space));

    firstBit = 1;
    reqSize = size;
    perSector = fs->SectorSize;

    while (reqSize != 0 && !error) 
    {
        bmap = fs->erasedBitmap;
        data = fs->erasedBitmap->bitmapData;
        dataEnd = fs->erasedBitmap->bitmapEnd;

        /* we're looking for "size" set bits in a row */
        for (; data <= dataEnd && reqSize > 0 && !error; data++) {
            for (idx = 0; idx < bitsPerByte && reqSize > 0 && !error; idx++) {
                if (*data & bitPositions[idx]) {
                    if (firstBit) { /* new first bit; save as start sector */
                        byteOffset = bmap->logSector * perSector
                                   + (data - (bmap->bitmapData));
                        space->start = byteOffset * bitsPerByte + idx;
                        space->length = (UINT32) 1;
                        firstBit = 0;
                    } else
                        space->length++;             /* it's adjacent */
                    reqSize--;
                    bmap->modified |= OT_MOD;           /* set dirty flag */
                    *data &= ~bitPositions[idx];        /* clear bit */
                } else {    /* hit used sector; start over and return sectors */
                    firstBit = 1;
                    if (reqSize != size)
                        error = returnSectors(fs,space,(UINT32) 0,(UINT32) 0,bmapErased);
                    reqSize = size;
                }
            }
        }

        /*  if not enough space in this bitmap sector to satifsy request, flush it
            (if necessary) and get next one */
        if (reqSize != 0) {
            if (bmap->modified & OT_MOD)
                error = ot_writeBitmap(fs,bmapErased);
            if (!error)
                error = ot_readBitmap(fs,bmap->logSector+1,bmapErased,0);
        }
    }
    if (reqSize != 0)   /* couldn't satisfy request */
        error = ENOSPC;
    if (error) {
        if (space)
            returnSectors(fs,space,(UINT32) 0,(UINT32) 0,bmapErased);
        kmem_free((caddr_t) space,(UINT32) sizeof(struct space),0);
        space = (struct space *) 0;
    } else
        fs->sectorsFree -= space->length;
    return space;
} /* end ot_getContiguous */

/********************************************************************/
/*  ot_getSpace                                                     */
/*      routine to get space on the disk to use                     */
/*                                                                  */
/*  input parameters:                                               */
/*      fs:         the otfs structure for the volume               */
/*      size:       the number of sectors needed                    */
/*      reqType:    type of space requested: contiguous or          */
/*                  nonContiguous                                   */ 
/*  returns:                                                        */
/*      0                   not enough space on device              */
/*      pointer to link(s)  success                                 */
/*                                                                  */
/********************************************************************/
struct space *ot_getSpace(register struct otfs *fs,
                          register UINT32 size,
                          enum spRequestType reqType)

{
    INT32                   idx;
    UINT32                  perSector,  /* number of bytes per sector used for bitmap */
                            byteInFile, /* byte offset into bitmap, not counting headers */
                            address,    /* calculated volume block address */
                            spRequired; /* # sectors to save for volume shutdown */
    struct space            *addrList;  /* for the returned information, in another form */
    struct bitmap           *bmap;
    struct otnode           *active,
                            *next;
    register INT32          error;
    register caddr_t        data,
                            dataEnd;

    /* if it's a WORM volume, save enough space to shut the volume down by making sure
        that we can flush all open files and system information */
    spRequired = 0;
    if (fs->MediaType <= (int) fsWORM) {
        for (active = fs->active; active; active = next) {
            next = active->next;
            if (active->flags & OT_MOD && !(active->flags & OT_INVAL))
                spRequired += ot_calcSpaceToFlush( fs, active->version);
        }
        /* 
        * ISSUE -
        * In the new NSR code, there is no base record. In calculating
        * space, include the freed space table, unallocated space table,
        * the partition integrity table.
        */
        spRequired += ot_calcSpaceToFlush(fs,fs->hwaterVersion);
        if (size > fs->sectorsFree-spRequired)
            return (struct space *) 0;
    }

    /* space is always allocated sequentially on WORM... */
    if (fs->MediaType <= (int) fsWORM || reqType == contiguous) {
        addrList = ot_getContiguous(fs,size);
        return addrList;
    }    

    /* if not enough available, erase cached unerased bitmap sector */
    if (size > fs->sectorsFree)
        if (ot_eraseUnerased(fs) || size > fs->sectorsFree)
            return (struct space *) 0;

    error = 0;
    addrList = (struct space *) 0;
    perSector = fs->SectorSize;

    /* check erased bitmap */
    while (size > 0 && !error) {
        bmap = fs->erasedBitmap;
        data = fs->erasedBitmap->bitmapData;
        dataEnd = fs->erasedBitmap->bitmapEnd;

        /*  look at each bit of each byte in the bitmap sector cached.  if it is
            available, calculate which sector it represents, put it in list to
            be returned, clear the bit in the bitmap and set the bitmap's dirty flag */
        for (; data <= dataEnd && size > 0 && !error; data++) {
            for (idx = 0; idx < bitsPerByte && size > 0 && !error; idx++) {
                if (*data & bitPositions[idx]) {
                    byteInFile = bmap->logSector * perSector + 
                                (data-(bmap->bitmapData));
                    address = byteInFile * bitsPerByte + idx;
                    error = ot_addToList(address,(UINT32) 1,&addrList,fs);
                    if (!error) {
                        size--;
                        fs->sectorsFree--;
                        *data &= ~bitPositions[idx];
                        fs->erasedBitmap->modified |= OT_MOD;
                    }
                }
            }
        }

        /* if we didn't fulfill the request, get the next bitmap sector and continue */
        if (size > 0 && !error) {
            if (fs->erasedBitmap->modified & OT_MOD)
                error = ot_writeBitmap(fs,bmapErased);            /* flush this sector */
            if (!error)
                error = ot_readBitmap(fs,bmap->logSector+1,bmapErased,0);
        }
    }
    if (size > 0) {                             /* couldn't satisfy request */
        returnSectors(fs,addrList,(UINT32) 0,(UINT32) 0,bmapErased);
        killSpaceList(&addrList);               /* free the (partial) list we created */
    }
    return size > 0 ? (struct space *) 0 : addrList;
} /* end ot_getSpace */

/********************************************************************/
/*  ot_putOnline                                                    */
/*      routine to bring a fs back online without calling otfs_mount*/
/*                                                                  */
/*  input parameters:                                               */
/*      fs:     the otfs structure for the volume                   */
/*                                                                  */
/*  returns:                                                        */
/*      0       success                                             */
/*      ENOMEM  memory allocation failure                           */
/*                                                                  */
/*  notes:                                                          */
/*      assumes ot_takeOffline was used to take the volume offline  */
/*          and merely does the reverse of that routine; note that  */
/*          only core code structures are restored here...          */
/*                                                                  */
/********************************************************************/
INT32 ot_putOnline(register struct otfs *fs)

{
    UINT32          realFreeCount;
    register INT32  error;

    error = 0;

    /* allocate otfs buffers */
    fs->iobuf = kmem_alloc((UINT32) (IOBUFSIZE * sizeof(char)),0);
    fs->dirbuf = kmem_alloc((UINT32) (DIRBUFSIZE * sizeof(char)),0);
    fs->erasedBitmap = (struct bitmap *) kmem_alloc((UINT32) sizeof(struct bitmap),0);
    fs->unerasedBitmap = (struct bitmap *) kmem_alloc((UINT32) sizeof(struct bitmap),0);
    if (!fs->iobuf || !fs->dirbuf ||
        !fs->erasedBitmap || !fs->unerasedBitmap)
        error = ENOMEM;

    if (!error && fs->MediaType > (int) fsWORM) {
        fs->erasedBitmap->bitmapHdr = kmem_alloc(fs->SectorSize,0);
        fs->unerasedBitmap->bitmapHdr = kmem_alloc(fs->SectorSize,0);
        if (!fs->erasedBitmap || !fs->unerasedBitmap)
            error = ENOMEM;
    }

    /*  get free space information.  this is a highwater mark on WORM, and bitmap
        information on rewritable */
    if (!error) {
        realFreeCount = fs->sectorsFree;
        if (fs->MediaType <= (int) fsWORM) {
            error = ot_rdHighwaterMark(fs);
            if (error)
                killSpaceList(&fs->highWater);
        } else {
            error = ot_readBitmap(fs,(UINT32) 0,bmapErased,1);
            if (!error)
                error = ot_readBitmap(fs,(UINT32) 0,bmapUnerased,1);
            if (error) {
                killBitmap(fs,bmapErased);
                killBitmap(fs,bmapUnerased);
            }
        }
        fs->sectorsFree = realFreeCount;
    }

    if (error) {
        if (fs->iobuf)
            kmem_free(fs->iobuf,(UINT32) IOBUFSIZE,0);
        if (fs->dirbuf)
            kmem_free(fs->dirbuf,(UINT32) DIRBUFSIZE,0);
    }
    return error;
} /* end ot_putOnline */

/********************************************************************/
/*  ot_rdHighwaterMark                                              */
/*      routine to get the highwater mark written to a WORM volume  */
/*                                                                  */
/*  input parameters:                                               */
/*      fs:     the otfs structure for the volume                   */
/*                                                                  */
/*  output parameters:                                              */
/*      fs:     fsp->highWater has changed                          */
/*                                                                  */
/*  returns:                                                        */
/*      0           success                                         */
/*      ENOMEM:     memory allocation failure                       */
/********************************************************************/
INT32 ot_rdHighwaterMark(register struct otfs *fs)

{
    UINT16                  numEntries;
    struct otnode           *otp;
    register INT32          error;
    struct UnallocSpEntry   *usptbl;
    struct gen_AD           *sp;

    /* ISSUE: 382 should come from otfs */
    error = ot_getOTNode(0,fs,"",fs->erasedAddr,LAST,VNON,(struct otnode *) 0,&otp);

    if (!error) { /* get the highwater mark sector and add entry to erased space list */
        usptbl = (struct UnallocSpEntry *)&otp->icbe;
        sp = USE_ADStart(usptbl);
        error = ot_addToList(sp->Location_LBN,sp->ExtentLength.bf.Length,&fs->highWater,fs);
    }

    if (error != 0 && error != EEOF)    /* a real error occurred reading the list */
        killSpaceList(&fs->highWater);
    if (error == 0 || error == EEOF)    /* success */
        fs->hwaterVersion = otp->version;
    if (otp) {
        otp->vnode.v_count--;
        otvn_inactive(&otp->vnode,(struct ucred *) 0);
    }
    return error == EEOF ? 0 : error;
} /* end ot_rdHighwaterMark */

/********************************************************************/
/*  ot_readBitmap                                                   */
/*      routine to read bitmap data.                                */
/*                                                                  */
/*  input parameters:                                               */
/*     fs:     the otfs structure for the volume                    */
/*     logSector: the logical sector number of bitmap (0,1,...)     */
/*     whichOne: specify erased (NSR unalloc) or unerased           */
/*     (NSR freed)                                                  */
/*     search: if 1, start at 1st logical sector and read until     */
/*             find a bit that is set.                              */
/*                                                                  */
/*  returns:                                                        */
/*      0       success                                             */
/********************************************************************/
INT32 ot_readBitmap(register struct otfs *fs,
                    UINT32 logSector,
                    enum bitmapType whichOne,
                    int search)

{
    UINT32              bmapSector;
    UINT32              SectorsInBmap; 
    UINT32              BytesInBmap; 
    UINT32              bsize; 
    UINT32              firstSector; 
    register INT32      error = 0;
    register int        fBitmap, found;
    register caddr_t    start, end;
    UCHAR8              isFirst;
    struct SpaceBitmapEntry *sb;
    struct bitmap       *bitmapp;

    found = 0;
    fBitmap = (whichOne == bmapErased);
    bsize = fs->SectorSize;

    bmapSector = (fBitmap ? fs->erasedAddr : fs->unerasedAddr);
    firstSector = bmapSector;
    
    if ( ! search )
        bmapSector += logSector;
    
    isFirst = (fBitmap ? bmapSector == fs->erasedAddr : bmapSector == fs->unerasedAddr);
    bitmapp = (fBitmap ? fs->erasedBitmap : fs->unerasedBitmap);

    /* space for bitmap already allocated */
    sb = (struct SpaceBitmapEntry *)bitmapp->bitmapHdr; 
    if ( isFirst ) {
        bitmapp->bitmapData = (char *)sb + sizeof(struct SpaceBitmapHdr);
    }
    else {
        bitmapp->bitmapData = bitmapp->bitmapHdr;
    }
    bitmapp->bitmapEnd = (char *)sb + fs->SectorSize - 1;

    /* read sectors in bitmap until we find next one to use */
    do
    {
        /* read in the space bitmap entry */
        if ( error = ot_read(fs,bmapSector,(caddr_t)sb,&bsize) ) {
            DPRINTF("ot_readBitmap: can't read erased bitmap\n");
            return error;
        }
#if !IBM
        if ( isFirst )  {
            ot_xlate(TAGID_SPACE_BMAP,FROM_OP,(caddr_t)sb,0,0);
        }
#endif
        if (search) {
            /* 
            * check each bit of the bitmap; we use the first sector 
            * with a bit set 
            */
            if (!error) {
                if ( isFirst ) {
                    start = (caddr_t)sb->bitmap; 
                    end = (caddr_t)sb + fs->SectorSize - 1;
                    if ( end - start > sb->hdr.N_Bytes ) {
                        end = start + sb->hdr.N_Bytes - 1;
                        bitmapp->bitmapEnd = end;
                    }
                    BytesInBmap = (sizeof(struct SpaceBitmapHdr) + sb->hdr.N_Bytes);
                    SectorsInBmap = BytesInBmap/fs->SectorSize 
                                  + ((BytesInBmap % fs->SectorSize) ? 1 : 0);
                    bitmapp->lastSector = bmapSector + SectorsInBmap - 1;
                    bitmapp->lastByte = (sb->hdr.N_Bytes + sizeof(struct SpaceBitmapHdr))
                                        % fs->SectorSize;
                    isFirst = 0;
                }
                else {
                    start = (caddr_t)sb;
                    if ( bmapSector != bitmapp->lastSector ) {
                        end = start + fs->SectorSize - 1;
                    }
                    else  {
                        end = start + bitmapp->lastByte - 1;
                    }
                }
                
                bitmapp->bitmapData = start;

                for ( ; start <= end && *start == byteClear; start++);

                found = (start <= end);
            }
        } else {
            found = 1;
        }
        if ( !found )
            bmapSector++;
    } while (!error && !found && bmapSector <= bitmapp->lastSector);

    bitmapp->logSector = bmapSector - firstSector;
    if ( bmapSector > bitmapp->lastSector ) {
        /* bmapSector went past the lastSector */
        bitmapp->logSector -= 1;
    }

    /* need to fix up bitmapEnd if this is the last sector */
    if ( bmapSector == bitmapp->lastSector )
        bitmapp->bitmapEnd = (char *)sb + bitmapp->lastByte - 1;
    
    /* did we try to go past the end ?? */
    if ( bmapSector > bitmapp->lastSector )
        error = EEOF;

    /*
    * return EEOF if not searching and past EOF, because ot_getContiguous
    * expects an error return  when it runs out of file 
    */

    return (((error == EEOF) && search) ? 0 : error);

} /* end ot_readBitmap */


/********************************************************************/
/*  ot_writeBitmap                                                  */
/********************************************************************/
INT32 ot_writeBitmap(register struct otfs *fs,
                     enum bitmapType whichOne)

{
    int                     fbitmap;
    INT32                   error = 0;
    struct bitmap           *bmap;
    UINT32                  sector;

    fbitmap = (whichOne == bmapErased);
    bmap = fbitmap ? fs->erasedBitmap : fs->unerasedBitmap;
    sector = fbitmap ? fs->erasedAddr : fs->unerasedAddr;
    sector += bmap->logSector;
    if (!error && (bmap->modified & OT_MOD)) {
#if !IBM
        if ( bmap->logSector == 0 ) {
            ot_xlate(TAGID_SPACE_BMAP,TO_OP,(caddr_t)bmap->bitmapHdr,0,0);
        }
#endif
        error = ot_write(fs,sector,bmap->bitmapHdr,fs->SectorSize);
        if (!error)
            bmap->modified &= ~OT_MOD;
    }
    return error;
} /* end ot_writeBitmap */

/********************************************************************/
/*  ot_wtHighwaterMark                                              */
/*      routine to write the highwater mark for a WORM volume       */
/*                                                                  */
/*  input parameters:                                               */
/*      fs:     the otfs structure for the volume                   */
/*                                                                  */
/*  returns:                                                        */
/*      0       success                                             */
/*      ENOMEM  memory allocation failure                           */
/********************************************************************/
INT32 ot_wtHighwaterMark(register struct otfs *fs)

{
    UINT32                address,
                          version;
    struct otnode         *fotp;
    register INT32        error;
    struct gen_AD         *ADp;

    /* if not modified, don't bother to flush it */
    if (!fs->highWater || !(fs->highWater->modified & OT_MOD))
        return 0;

    error = ot_getOTNode( 0, fs, "", fs->erasedAddr, LAST, VNON, (struct otnode *) 0, &fotp);
    if (!fotp)
        return ENOMEM;

    /*
     * Allocate space for icb entry.
     *
     * NOTE that this must be done separately from the writing of the
     * icb in order for the space required by the icbe itself to be taken
     * out of the erased bitmap before writing it to disk. This precludes the
     * more conventional writing of this through ot_ICBRdWr(). Therefore,
     * the structure of this function should reflect very closely the
     * structure of ot_ICBRdWr().
     */
    bzero(fs->iobuf,IOBUFSIZE);
    error = ot_searchLast( fs, fs->erasedAddr, &address, fs->iobuf, UIO_WRITE, &version);

    if (!error) {
        ADp = fotp->ADL.beginADs;  /* Single AD entry! */
        ADp->ExtentLength.bf.Type = E_UNALLOCATED;
        ADp->InfoLength  =
            ADp->RecordedLength =
                ADp->ExtentLength.bf.Length = fs->highWater->length-1;
        ADp->Location_LBN = fs->highWater->start;
        ADp->Location_PartNo = 0;

        /* update tag location, checksum */
        FillTag(fotp->icbe.ge.sTag.uTagID,
                fotp->icbe.ge.sTag.uDescriptorVersion,
                fotp->icbe.ge.sTag.uTagSerialNum,
                0,0,address,(struct Tag *)&fotp->icbe.ge.sTag);

        fotp->version++;

        bzero(fs->iobuf,(INT32) IOBUFSIZE);
        ot_copy( (caddr_t) &fotp->icbe, fs->iobuf,fs->SectorSize);
#if !IBM
        ot_xlate(TAGID_UNALLOC_SP_ENTRY,TO_OP,
                 (caddr_t)&fotp->icbe,fs->SectorSize,0);
#endif
        if (!error) {
            error=ot_write(fs,address,(caddr_t)&fotp->icbe,fs->SectorSize);
        }
        if (!error) {
            fs->highWater->modified &= ~OT_MOD;
            fs->hwaterVersion = fotp->version;
        }
    }

    fotp->vnode.v_count--;
    otvn_inactive( &fotp->vnode, (struct ucred *) 0);

    return error;
} /* end ot_wtHighwaterMark */


/********************************************************************/
/*  killBitmap                                                      */
/********************************************************************/
void killBitmap(register struct otfs *fs,
                enum bitmapType whichMap)

{
    caddr_t                 data;
    register struct bitmap  *bmap;

    data = (caddr_t) 0;
    bmap = whichMap == bmapErased ? fs->erasedBitmap : fs->unerasedBitmap;
    if (bmap)
        data = bmap->bitmapHdr;

    if (data)
        kmem_free( data, fs->SectorSize, 0);
    if (bmap)
        kmem_free( (caddr_t) bmap, (UINT32) sizeof(struct bitmap), 0);
    if (whichMap == bmapErased)
        fs->erasedBitmap = (struct bitmap *) 0;
    else
        fs->unerasedBitmap = (struct bitmap *) 0;
} /* end killBitmap */

/********************************************************************/
/*  killSpaceList                                                   */
/*      free an entire space list                                   */
/*                                                                  */
/*  input parameters:                                               */
/*      list:       head of list to be freed                        */
/*                                                                  */
/********************************************************************/
void killSpaceList(struct space **list)

{
    struct space            *next;
    register struct space   *s = *list;

    while (s) {
        next = s->next;
        kmem_free((caddr_t) s,(UINT32) sizeof(struct space),0);
        s = next;
    }
    *list = (struct space *) 0;
} /* end killSpaceList */

/********************************************************************/
/* power                                                            */
/********************************************************************/
UINT32 power(UINT32 x,
             UINT32 n)

{
    UINT32  i,
            p;

    p = 1;
    for (i = 1; i <= n; i++)
        p *= x;
    return p;
} /* power */

/********************************************************************/
/*  returnSectors                                                   */
/********************************************************************/
INT32 returnSectors(register struct otfs *fs,
                    struct space *list,
                    UINT32 start,
                    UINT32 length,
                    enum bitmapType bmType)


{
    UINT32                  perSector,  /* # sectors represented in 1 bitmap sector */
                            byteOffset, /* byte in sector for a given volume address */
                            bitOffset;  /* bit within a byte for given volume address */
    struct space            dsList; 
    caddr_t                 bmData;
    register INT32          error;
    register UINT32         addr,       /* address counter */
                            logSector;  /* sector of bitmap for given volume address */
    register struct space   *tmp;
    register struct bitmap  *bmap;

    if (fs->MediaType <= (int) fsWORM)
        return 0;

    error = 0;
    bmap = bmType == bmapErased ? fs->erasedBitmap : fs->unerasedBitmap;
    bmData = bmap->bitmapData;
    perSector = fs->SectorSize * bitsPerByte;
    if (!list) {
        bzero((caddr_t) &dsList,(INT32) sizeof(struct space));
        dsList.start = start;
        dsList.length = length;
    }

    /*  for each sector in each link of the list, calculate which logical sector
        of the bitmap it belongs to.  if the sector does not belong in the
        bitmap sector we have cached, flush the sector we have (if needed),
        and read the bitmap sector we need. */
    for (tmp = list ? list : &dsList; tmp; tmp = tmp->next) {
        for (addr = tmp->start; addr < tmp->start+tmp->length; addr++) {
            logSector = (addr+sizeof(struct SpaceBitmapHdr))/perSector;
            if (logSector != bmap->logSector) {
                if (bmap->modified & OT_MOD)
                    error = ot_writeBitmap(fs,bmType);
                if (!error)
                    error = ot_readBitmap(fs,logSector,bmType,0);   
            }
            /*  now that we have the appropriate sector, reset the bit which
                represents this address and set bitmap dirty bit */
            byteOffset = (addr - (bmap->logSector * perSector)) / bitsPerByte;

            /* bitOffset = the address - first address this byte represents */
            bitOffset = addr - ((fs->SectorSize * logSector) + byteOffset) * bitsPerByte;
            bmData[byteOffset] |= bitPositions[bitOffset];
            bmap->modified |= OT_MOD;
        }
        /* increment sectors free counter if we returned sectors to the erased bitmap */
        if (bmType == bmapErased && !error)
            fs->sectorsFree += list ? tmp->length : length;
    }

    return error;
} /* end returnSectors */

/********************************************************************/
/*  uioInit                                                         */
/*      initialize a uio structure                                  */
/*                                                                  */
/*  input parameters:                                               */
/*      uiop:       the structure to initialize                     */
/*      iovecp:     the iovec structure to be used                  */
/*      iobuffer:   for I/O                                         */
/*      offset:     byte offset; where to start read/write          */
/*      length:     residual; how much to read/write                */
/*                                                                  */
/*  returns:                                                        */
/*      none                                                        */
/*                                                                  */
/********************************************************************/
void uioInit(register struct uio *uiop,
             register struct iovec *iovecp,
             caddr_t iobuffer,
             INT32 offset,
             INT32 length)

{
    iovecp->iov_base = iobuffer;
    iovecp->iov_len = length;
    uiop->uio_iov = iovecp;
    uiop->uio_iovcnt = 1;
    uiop->uio_offset = offset;
    uiop->uio_seg = UIOSEG_KERNEL;
#if (SUN || MACINTOSH || IBM)
    uiop->uio_fmode = 0;
#endif
    uiop->uio_resid = length;
} /* end uioInit */

/********************************************************************/
/*  vattrInit                                                       */
/*      initialize a vattr structure                                */
/*                                                                  */
/*  input parameters:                                               */
/*      va:     the structure to initialize                         */
/********************************************************************/
void vattrInit( struct vattr *va)

{
    va->va_nlink = -1;
    va->va_blocksize = -1;
    va->va_rdev = -1;
    va->va_blocks = -1;
    va->va_fsid = -1;
    va->va_nodeid = -1;
    va->va_type = VREG;
} /* end vattrInit */
